# Copyright (C) 2020-2022 Xilinx, Inc.
# Copyright (C) 2022-2025, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Collection of default build steps for building and verifying a dataflow
accelerator from an ONNX model.
"""

import json
import numpy as np
import os
import shutil
from copy import deepcopy
from functools import partial
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.custom_op.registry import getCustomOp
from qonnx.transformation.bipolar_to_xnor import ConvertBipolarMatMulToXnorPopcount
from qonnx.transformation.fold_constants import FoldConstants
from qonnx.transformation.general import (
    ApplyConfig,
    GiveReadableTensorNames,
    GiveUniqueNodeNames,
    RemoveStaticGraphInputs,
    RemoveUnusedTensors,
    SortGraph,
)
from qonnx.transformation.infer_data_layouts import InferDataLayouts
from qonnx.transformation.infer_datatypes import InferDataTypes
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.transformation.lower_convs_to_matmul import LowerConvsToMatMul
from qonnx.util.cleanup import cleanup_model
from qonnx.util.config import extract_model_config_to_json
from shutil import copy

import finn.transformation.fpgadataflow.convert_to_hw_layers as to_hw
import finn.transformation.streamline.absorb as absorb
from finn.analysis.fpgadataflow.dataflow_performance import dataflow_performance
from finn.analysis.fpgadataflow.exp_cycles_per_layer import exp_cycles_per_layer
from finn.analysis.fpgadataflow.hls_synth_res_estimation import hls_synth_res_estimation
from finn.analysis.fpgadataflow.op_and_param_counts import aggregate_dict_keys, op_and_param_counts
from finn.analysis.fpgadataflow.post_synth_res import post_synth_res
from finn.analysis.fpgadataflow.res_estimation import res_estimation, res_estimation_complete
from finn.analysis.fpgadataflow.unsupported_layers import unsupported_layers
from finn.builder.build_dataflow_config import (
    DataflowBuildConfig,
    DataflowOutputType,
    ShellFlowType,
    VerificationStepType,
)
from finn.core.onnx_exec import execute_onnx
from finn.core.rtlsim_exec import rtlsim_exec
from finn.transformation.fpgadataflow.annotate_cycles import AnnotateCycles
from finn.transformation.fpgadataflow.compile_cppsim import CompileCppSim
from finn.transformation.fpgadataflow.create_dataflow_partition import CreateDataflowPartition
from finn.transformation.fpgadataflow.create_stitched_ip import CreateStitchedIP
from finn.transformation.fpgadataflow.derive_characteristic import (
    DeriveCharacteristic,
    DeriveFIFOSizes,
)
from finn.transformation.fpgadataflow.hlssynth_ip import HLSSynthIP
from finn.transformation.fpgadataflow.insert_dwc import InsertDWC
from finn.transformation.fpgadataflow.insert_fifo import InsertFIFO
from finn.transformation.fpgadataflow.insert_tlastmarker import InsertTLastMarker
from finn.transformation.fpgadataflow.make_driver import (
    MakeCPPDriver,
    MakePYNQDriverInstrumentation,
    MakePYNQDriverIODMA,
    update_bitfile_path_after_copy,
)
from finn.transformation.fpgadataflow.make_zynq_proj import ZynqBuild
from finn.transformation.fpgadataflow.minimize_accumulator_width import MinimizeAccumulatorWidth
from finn.transformation.fpgadataflow.minimize_weight_bit_width import MinimizeWeightBitWidth
from finn.transformation.fpgadataflow.prepare_cppsim import PrepareCppSim
from finn.transformation.fpgadataflow.prepare_ip import PrepareIP
from finn.transformation.fpgadataflow.prepare_rtlsim import PrepareRTLSim
from finn.transformation.fpgadataflow.replace_verilog_relpaths import ReplaceVerilogRelPaths
from finn.transformation.fpgadataflow.set_exec_mode import SetExecMode
from finn.transformation.fpgadataflow.set_fifo_depths import (
    InsertAndSetFIFODepths,
    RemoveShallowFIFOs,
    SplitLargeFIFOs,
    xsi_fifosim,
)
from finn.transformation.fpgadataflow.set_folding import SetFolding
from finn.transformation.fpgadataflow.specialize_layers import SpecializeLayers
from finn.transformation.fpgadataflow.synth_ooc import SynthOutOfContext
from finn.transformation.fpgadataflow.vitis_build import VitisBuild
from finn.transformation.fpgadataflow.vivado_power_estimation import VivadoPowerEstimation
from finn.transformation.move_reshape import RemoveCNVtoFCFlatten
from finn.transformation.qonnx.convert_qonnx_to_finn import ConvertQONNXtoFINN
from finn.transformation.qonnx.quant_act_to_multithreshold import default_filter_function_generator
from finn.transformation.streamline import Streamline
from finn.transformation.streamline.reorder import MakeMaxPoolNHWC
from finn.transformation.streamline.round_thresholds import RoundAndClipThresholds
from finn.util.basic import get_liveness_threshold_cycles, get_rtlsim_trace_depth
from finn.util.exception import FINNUserError
from finn.util.execution import execute_parent
from finn.util.logging import log


def verify_step(
    model: ModelWrapper,
    cfg: DataflowBuildConfig,
    step_name: str,
    need_parent: bool,
    rtlsim_pre_hook=None,
):
    """Verify a build step by running simulation and comparing results.

    Args:
        model: The ONNX model to verify
        cfg: Build configuration object
        step_name: Name of the build step being verified
        need_parent: Whether parent model execution is needed for comparison
        rtlsim_pre_hook: Optional pre-hook function for RTL simulation
    """
    log.info(f"Running verification for {step_name}")
    verify_out_dir = cfg.output_dir + "/verification_output"
    intermediate_models_dir = cfg.output_dir + "/intermediate_models"
    # Ensure tensor names are sorted and readable for easier debugging
    model = model.transform(SortGraph())
    model = model.transform(GiveUniqueNodeNames())
    model = model.transform(GiveReadableTensorNames())
    os.makedirs(verify_out_dir, exist_ok=True)
    (in_npy_all, exp_out_npy_all) = cfg._resolve_verification_io_pair()
    bsize_in = in_npy_all.shape[0]
    bsize_out = exp_out_npy_all.shape[0]
    assert bsize_in == bsize_out, "Batch sizes don't match for verification IO pair"
    all_res = True
    for b in range(bsize_in):
        in_npy = np.expand_dims(in_npy_all[b], axis=0)
        exp_out_npy = np.expand_dims(exp_out_npy_all[b], axis=0)
        if need_parent:
            assert cfg.save_intermediate_models, "Enable save_intermediate_models for verification"
            parent_model_fn = intermediate_models_dir + "/dataflow_parent.onnx"
            child_model_fn = intermediate_models_dir + "/verify_%s.onnx" % step_name
            model.save(child_model_fn)
            parent_model = ModelWrapper(parent_model_fn)
            out_tensor_name = parent_model.graph.output[0].name
            exp_ishape = parent_model.get_tensor_shape(parent_model.graph.input[0].name)
            if in_npy.shape != exp_ishape:
                log.warning(
                    f"Verification input has shape {in_npy.shape} while model expects {exp_ishape}"
                )
                log.info("Attempting to force model shape on verification input")
                in_npy = in_npy.reshape(exp_ishape)
            out_dict = execute_parent(parent_model_fn, child_model_fn, in_npy, return_full_ctx=True)
            out_npy = out_dict[out_tensor_name]
        else:
            inp_tensor_name = model.graph.input[0].name
            out_tensor_name = model.graph.output[0].name
            exp_ishape = model.get_tensor_shape(inp_tensor_name)
            if in_npy.shape != exp_ishape:
                log.warning(
                    f"Verification input has shape {in_npy.shape} while model expects {exp_ishape}"
                )
                log.info("Attempting to force model shape on verification input")
                in_npy = in_npy.reshape(exp_ishape)
            inp_dict = {inp_tensor_name: in_npy}
            if rtlsim_pre_hook is not None:
                out_dict = rtlsim_exec(model, inp_dict, pre_hook=rtlsim_pre_hook)
            else:
                out_dict = execute_onnx(model, inp_dict, True)
            out_npy = out_dict[out_tensor_name]
        exp_oshape = exp_out_npy.shape
        if out_npy.shape != exp_oshape:
            log.warning(
                f"Verification input has shape {exp_oshape} while model expects {out_npy.shape}"
            )
            log.info("Attempting to force model shape on verification input")
            out_npy = out_npy.reshape(exp_oshape)

        # Check 1: Element-wise closeness between output and expected output
        res1 = np.isclose(
            out_npy, exp_out_npy, atol=cfg.verification_atol, rtol=cfg.verification_rtol
        ).all()
        # Check 2 and 3: Mean absolute and relative error over all output elements
        num_elements = out_npy.size
        abs_error = np.abs(out_npy - exp_out_npy)
        # Avoid division by zero for relative error
        exp_out_npy_safe = np.where(exp_out_npy == 0, np.finfo(float).eps, exp_out_npy)
        rel_error = np.abs((out_npy - exp_out_npy) / exp_out_npy_safe)
        res2 = np.mean(abs_error) <= cfg.verification_mean_atol
        res3 = np.mean(rel_error) <= cfg.verification_mean_rtol

        res = res1 and res2 and res3
        all_res = all_res and res
        res_to_str = {True: "SUCCESS", False: "FAIL"}
        res_str = res_to_str[res]
        if cfg.verify_save_full_context:
            verification_output_fn = os.path.join(
                verify_out_dir, f"verify_{step_name}_{b}_{res_str}.npz"
            )
            np.savez(verification_output_fn, **out_dict)

            # Log tensor statistics for debugging (only output tensors, in topological order)
            tensors_to_log = ["global_in"]
            if need_parent:
                for node in parent_model.graph.node:
                    for output in node.output:
                        tensors_to_log.append(output)
                sdp_node = parent_model.get_nodes_by_op_type("StreamingDataflowPartition")[0]
                sdp_prefix = sdp_node.name + "_"
            else:
                sdp_prefix = ""
            for node in model.graph.node:
                for output in node.output:
                    tensors_to_log.append(sdp_prefix + output)

            tensor_stats = []
            for key in tensors_to_log:
                if key in out_dict:
                    stat_dict = {
                        "tensor": key,
                        "shape": list(out_dict[key].shape),
                        "mean": float(np.mean(out_dict[key])),
                        "std": float(np.std(out_dict[key])),
                        "min": float(np.min(out_dict[key])),
                        "max": float(np.max(out_dict[key])),
                    }
                    tensor_stats.append(stat_dict)

            # Write tensor statistics in compact human-readable table format
            with open(
                os.path.join(verify_out_dir, f"verify_{step_name}_{b}_{res_str}_stats.txt"), "w"
            ) as f:
                # Write header
                f.write(
                    f"{'Tensor':<40} {'Shape':<20} {'Mean':<12} "
                    f"{'Std':<12} {'Min':<12} {'Max':<12}\n"
                )
                f.write("-" * 108 + "\n")

                # Write data rows
                for stat in tensor_stats:
                    # Shorten/truncate long names and shapes
                    tensor_name = stat["tensor"].replace("GenericPartition", "GP")[:39]
                    shape_str = str(stat["shape"])[:19]
                    f.write(
                        f"{tensor_name:<40} {shape_str:<20} {stat['mean']:<12.6f} "
                        f"{stat['std']:<12.6f} {stat['min']:<12.6f} {stat['max']:<12.6f}\n"
                    )

                # Add output error analysis
                f.write("\n" + "=" * 108 + "\n")
                f.write("OUTPUT ERROR ANALYSIS\n")
                f.write("=" * 108 + "\n")

                f.write(f"Number of elements:           {num_elements}\n")
                f.write(f"Min absolute error:           {np.min(abs_error):.6e}\n")
                f.write(f"Max absolute error:           {np.max(abs_error):.6e}\n")
                f.write(f"Mean absolute error:          {np.mean(abs_error):.6e}\n")
                f.write(f"Min relative error:           {np.min(rel_error):.6e}\n")
                f.write(f"Max relative error:           {np.max(rel_error):.6e}\n")
                f.write(f"Mean relative error:          {np.mean(rel_error):.6e}\n")
                f.write(
                    f"Tolerance per element:        atol={cfg.verification_atol:.6e} + "
                    f"rtol={cfg.verification_rtol:.6e}\n"
                )
                f.write(f"Tolerance for mean abs. err:  {cfg.verification_mean_atol:.6e}\n")
                f.write(f"Tolerance for mean rel. err:  {cfg.verification_mean_rtol:.6e}\n")
                f.write(f"Verification result:          {res_str}\n")
        else:
            verification_output_fn = os.path.join(
                verify_out_dir, f"verify_{step_name}_{b}_{res_str}.npy"
            )
            np.save(verification_output_fn, out_npy)

        if cfg.verify_save_rtlsim_waveforms:
            wdb_path = model.get_metadata_prop("rtlsim_trace")
            if wdb_path is not None and os.path.isfile(wdb_path):
                new_wdb_path = wdb_path.replace(".wdb", "_%d.wdb" % b)
                shutil.move(wdb_path, new_wdb_path)

    log.info(f"Verification for {step_name} : {res_to_str[all_res]}")


def prepare_for_stitched_ip_rtlsim(verify_model, cfg):
    """Prepare model for stitched IP RTL simulation.

    Switches implementation styles from Vivado components to RTL where needed
    and ensures proper configuration for RTL simulation.

    Args:
        verify_model: The model to prepare for RTL simulation
        cfg: Build configuration object
    """
    if not cfg.rtlsim_use_vivado_comps:
        need_restitch = False
        # switch impl_style=vivado components to rtl
        # StreamingFIFO must have impl_style=rtl
        for fifo_layer in verify_model.get_nodes_by_op_type("StreamingFIFO_rtl"):
            inst = getCustomOp(fifo_layer)
            if inst.get_nodeattr("impl_style") != "rtl":
                inst.set_nodeattr("impl_style", "rtl")
                inst.set_nodeattr("code_gen_dir_ipgen", "")
                inst.set_nodeattr("ipgen_path", "")
                need_restitch = True
        # if we've made alterations to the model, need to do some re-prep
        if need_restitch:
            log.info("Need to regen/re-stitch some IP for STITCHED_IP_RTLSIM")
            verify_model = verify_model.transform(
                PrepareIP(cfg._resolve_fpga_part(), cfg._resolve_hls_clk_period())
            )
            verify_model = verify_model.transform(HLSSynthIP())
            verify_model = verify_model.transform(
                CreateStitchedIP(
                    cfg._resolve_fpga_part(),
                    cfg.synth_clk_period_ns,
                    vitis=False,
                )
            )
    else:
        log.info("rtlsim_use_vivado_comps is enabled, may yield incorrect results")

    # set top-level prop for stitched-ip rtlsim and launch
    verify_model.set_metadata_prop("exec_mode", "rtlsim")
    # TODO make configurable
    # verify_model.set_metadata_prop("rtlsim_trace", "trace.vcd")
    return verify_model


def step_qonnx_to_finn(model: ModelWrapper, cfg: DataflowBuildConfig):
    """
    This step will only execute if QONNX nodes are found.
    These include the following op_types: "Quant" , "Trunc" and "BinaryQuant".
    If such nodes are found the step will run the tidy-up step from QONNX
    and then convert the QONNX model to the FINN-ONNX dialect.
    """
    # Check if any QONNX nodes exist, i.e. BipolarQuant, BinaryQuant, Quant or Trunc
    q_count = 0
    for op_type in ["BipolarQuant", "BinaryQuant", "Quant", "Trunc"]:
        q_count += len(model.get_nodes_by_op_type(op_type))
    if q_count == 0:
        return model

    # QONNX cleanup
    model = cleanup_model(model)
    # QONNX to FINN-ONNX
    model = model.transform(
        ConvertQONNXtoFINN(
            filter_function=default_filter_function_generator(
                max_multithreshold_bit_width=cfg.max_multithreshold_bit_width
            )
        )
    )

    if VerificationStepType.QONNX_TO_FINN_PYTHON in cfg._resolve_verification_steps():
        verify_step(model, cfg, "finn_onnx_python", need_parent=False)

    return model


def step_tidy_up(model: ModelWrapper, cfg: DataflowBuildConfig):
    """Run the tidy-up step on given model. This includes shape and datatype
    inference, constant folding, and giving nodes and tensors better names.
    """

    model = model.transform(InferShapes())
    model = model.transform(FoldConstants())
    model = model.transform(GiveUniqueNodeNames())
    model = model.transform(GiveReadableTensorNames())
    model = model.transform(InferDataTypes())
    model = model.transform(RemoveStaticGraphInputs())

    if VerificationStepType.TIDY_UP_PYTHON in cfg._resolve_verification_steps():
        verify_step(model, cfg, "initial_python", need_parent=False)

    return model


def step_streamline(model: ModelWrapper, cfg: DataflowBuildConfig):
    """Run streamlining on given model. Streamlining involves moving floating point
    scale/shift parameters around, collapsing adjacent ones into a single parameter,
    then absorbing the scale/shift into the following `MultiThreshold` node.
    Streamlining requires careful topology design and cannot be applied to all
    topologies.
    """

    model = model.transform(absorb.AbsorbSignBiasIntoMultiThreshold())
    model = model.transform(Streamline())
    need_lowering = len(model.get_nodes_by_op_type("Conv")) > 0
    if need_lowering:
        model = model.transform(LowerConvsToMatMul())
        model = model.transform(MakeMaxPoolNHWC())
        model = model.transform(absorb.AbsorbTransposeIntoMultiThreshold())
        model = model.transform(MakeMaxPoolNHWC())
        model = model.transform(absorb.AbsorbConsecutiveTransposes())
    model = model.transform(ConvertBipolarMatMulToXnorPopcount())
    model = model.transform(Streamline())
    # absorb final add-mul nodes into TopK
    model = model.transform(absorb.AbsorbScalarMulAddIntoTopK())
    model = model.transform(InferDataLayouts())
    model = model.transform(RemoveUnusedTensors())

    if VerificationStepType.STREAMLINED_PYTHON in cfg._resolve_verification_steps():
        verify_step(model, cfg, "streamlined_python", need_parent=False)

    return model


def step_convert_to_hw(model: ModelWrapper, cfg: DataflowBuildConfig):
    """Convert eligible nodes to `HWCustomOp` subclasses that represent HW
    layers. Which nodes and particular configurations can be converted to HW
    is limited, see the source code of the `convert_to_hw` module for more.
    In the end am empty json file is created which can be used to set user specific
    preferred implementation styles for each node."""

    if cfg.standalone_thresholds:
        # doing this first causes all threshold layers to be standalone
        model = model.transform(to_hw.InferThresholdingLayer())
    # needed for bipolar MatMul layers
    model = model.transform(to_hw.InferBinaryMatrixVectorActivation())
    # needed for non-bipolar MatMul layers
    model = model.transform(to_hw.InferQuantizedMatrixVectorActivation())
    # TopK to LabelSelect
    model = model.transform(to_hw.InferLabelSelectLayer())
    # input quantization (if any) as standalone threshold
    model = model.transform(to_hw.InferThresholdingLayer())
    # needed for convolutions -- TODO always exec?
    need_conv = len(model.get_nodes_by_op_type("Im2Col")) > 0
    if need_conv:
        model = model.transform(to_hw.InferPool())
        model = model.transform(to_hw.InferConvInpGen())
        model = model.transform(RemoveCNVtoFCFlatten())
    # get rid of Tranpose -> Tranpose identity seq
    model = model.transform(absorb.AbsorbConsecutiveTransposes())
    model = model.transform(GiveUniqueNodeNames())
    model = model.transform(InferDataLayouts())
    model = model.transform(InferDataTypes())
    model = model.transform(InferShapes())
    model = model.transform(RoundAndClipThresholds())
    model = model.transform(to_hw.InferReshape())
    return model


def step_create_dataflow_partition(model: ModelWrapper, cfg: DataflowBuildConfig):
    """Separate consecutive groups of HWCustomOp nodes into StreamingDataflowPartition
    nodes, which point to a separate ONNX file. Dataflow accelerator synthesis
    can only be performed on those HWCustomOp sub-graphs."""

    unmapped_layers = [
        node.name
        for node in model.graph.node
        if not node.domain.startswith("finn.custom_op.fpgadataflow")
    ]
    # Check if there are unsupported layers somewhere between supported layers
    # This would cause a "cyclic-free graph partitioning violated" error otherwise
    results = model.analysis(unsupported_layers)
    if not results[0]:
        raise FINNUserError(
            f"Unsupported/unmapped layer(s) found in between FINN operators, "
            f"starting with node {results[1].name}. "
            "Complete list of unmapped nodes: " + ", ".join(unmapped_layers)
        )

    # Warn if unsupported layers remain at the start or end of the graph
    if unmapped_layers:
        log.warning(
            "The following nodes at the start/end of the graph will not be mapped to the "
            "accelerator, so they will need to be implemented manually (e.g., in software): "
            + ", ".join(unmapped_layers)
        )

    parent_model = model.transform(
        CreateDataflowPartition(
            partition_model_dir=cfg.output_dir + "/intermediate_models/supported_op_partitions"
        )
    )
    sdp_nodes = parent_model.get_nodes_by_op_type("StreamingDataflowPartition")
    assert len(sdp_nodes) == 1, "Only a single StreamingDataflowPartition supported."
    sdp_node = sdp_nodes[0]
    sdp_node = getCustomOp(sdp_node)
    dataflow_model_filename = sdp_node.get_nodeattr("model")
    if cfg.save_intermediate_models:
        parent_model.save(cfg.output_dir + "/intermediate_models/dataflow_parent.onnx")
    model = ModelWrapper(dataflow_model_filename)

    # create a configuration json file that can be used to set the specialize layer config
    attrs = [
        "preferred_impl_style",
    ]
    extract_model_config_to_json(
        model, cfg.output_dir + "/template_specialize_layers_config.json", attrs
    )

    return model


def step_specialize_layers(model: ModelWrapper, cfg: DataflowBuildConfig):
    """Convert HW nodes to either an HLS or RTL variant of the node. HW nodes
    get converted either based on pre-determined rules (details can be found
    in `specialize_layers` source code) or the user provides a configuration file
    which contains the desired setting. If the user preference cannot be fulfilled,
    a warning will be printed and the implementation style will be set to a default."""

    if cfg.specialize_layers_config_file is not None:
        model = model.transform(GiveUniqueNodeNames())
        model = model.transform(ApplyConfig(cfg.specialize_layers_config_file))
    model = model.transform(SpecializeLayers(cfg._resolve_fpga_part()))
    model = model.transform(InferShapes())
    model = model.transform(InferDataTypes())
    return model


def step_target_fps_parallelization(model: ModelWrapper, cfg: DataflowBuildConfig):
    """If target_fps was specified, use the SetFolding transformation to determine
    parallelization attributes. The auto-generated config will be saved under
    auto_folding_config.json under the outputs, which can serve as a basis for
    customizing the folding factors further."""

    target_cycles_per_frame = cfg._resolve_cycles_per_frame()
    if target_cycles_per_frame is not None:
        model = model.transform(
            SetFolding(
                target_cycles_per_frame,
                mvau_wwidth_max=cfg.mvau_wwidth_max,
                two_pass_relaxation=cfg.folding_two_pass_relaxation,
            )
        )
        # extract the suggested configuration and save it as json
        hw_attrs = [
            "PE",
            "SIMD",
            "parallel_window",
            "ram_style",
            "resType",
            "mem_mode",
            "runtime_writeable_weights",
            "depth_trigger_uram",
            "depth_trigger_bram",
        ]
        extract_model_config_to_json(
            model, cfg.output_dir + "/report/auto_folding_config.json", hw_attrs
        )

    return model


def step_apply_folding_config(model: ModelWrapper, cfg: DataflowBuildConfig):
    """Apply the folding configuration file onto the model to set folding (parallelization)
    and other attributes, if config file is specified."""

    if cfg.folding_config_file is not None:
        model = model.transform(GiveUniqueNodeNames())
        model = model.transform(ApplyConfig(cfg.folding_config_file))

    if VerificationStepType.FOLDED_HLS_CPPSIM in cfg._resolve_verification_steps():
        # prepare cppsim
        model = model.transform(PrepareCppSim())
        model = model.transform(CompileCppSim())
        model = model.transform(SetExecMode("cppsim"))
        verify_step(model, cfg, "folded_hls_cppsim", need_parent=True)
    return model


def step_generate_estimate_reports(model: ModelWrapper, cfg: DataflowBuildConfig):
    "Generate per-layer resource and cycle estimates using analytical models."

    if DataflowOutputType.ESTIMATE_REPORTS in cfg.generate_outputs:
        report_dir = cfg.output_dir + "/report"
        os.makedirs(report_dir, exist_ok=True)
        ops_and_params = model.analysis(op_and_param_counts)
        with open(report_dir + "/op_and_param_counts.json", "w") as f:
            json.dump(ops_and_params, f, indent=2)
        estimate_layer_cycles = model.analysis(exp_cycles_per_layer)
        with open(report_dir + "/estimate_layer_cycles.json", "w") as f:
            json.dump(estimate_layer_cycles, f, indent=2)
        estimate_layer_resources = model.analysis(
            partial(res_estimation, fpgapart=cfg._resolve_fpga_part())
        )
        estimate_layer_resources["total"] = aggregate_dict_keys(estimate_layer_resources)
        with open(report_dir + "/estimate_layer_resources.json", "w") as f:
            json.dump(estimate_layer_resources, f, indent=2)
        estimate_layer_resources_complete = model.analysis(
            partial(res_estimation_complete, fpgapart=cfg._resolve_fpga_part())
        )
        with open(report_dir + "/estimate_layer_config_alternatives.json", "w") as f:
            json.dump(estimate_layer_resources_complete, f, indent=2)
        # need to call AnnotateCycles before dataflow_performance
        model = model.transform(AnnotateCycles())
        estimate_network_performance = model.analysis(dataflow_performance)
        # add some more metrics to estimated performance
        n_clock_cycles_per_sec = (10**9) / cfg.synth_clk_period_ns
        est_fps = n_clock_cycles_per_sec / estimate_network_performance["max_cycles"]
        estimate_network_performance["estimated_throughput_fps"] = est_fps
        est_latency_ns = (
            estimate_network_performance["critical_path_cycles"] * cfg.synth_clk_period_ns
        )
        estimate_network_performance["estimated_latency_ns"] = est_latency_ns
        with open(report_dir + "/estimate_network_performance.json", "w") as f:
            json.dump(estimate_network_performance, f, indent=2)
    return model


def step_minimize_bit_width(model: ModelWrapper, cfg: DataflowBuildConfig):
    """Tighten the weight and accumulator bit widths for each layer."""
    if cfg.minimize_bit_width:
        model = model.transform(MinimizeWeightBitWidth())
        model = model.transform(MinimizeAccumulatorWidth())
        model = model.transform(RoundAndClipThresholds())
        # make sure the changed datatypes are propagated through the network
        model = model.transform(InferDataTypes())
    return model


def step_hw_codegen(model: ModelWrapper, cfg: DataflowBuildConfig):
    """Generate Vitis HLS code to prepare HLSBackend nodes for IP generation.
    And fills RTL templates for RTLBackend nodes."""

    model = model.transform(PrepareIP(cfg._resolve_fpga_part(), cfg._resolve_hls_clk_period()))
    return model


def step_hw_ipgen(model: ModelWrapper, cfg: DataflowBuildConfig):
    """Run Vitis HLS synthesis on generated code for HLSBackend nodes,
    in order to generate IP blocks. For RTL nodes this step does not do anything."""

    model = model.transform(HLSSynthIP())
    model = model.transform(ReplaceVerilogRelPaths())
    report_dir = cfg.output_dir + "/report"
    os.makedirs(report_dir, exist_ok=True)
    estimate_layer_resources_hls = model.analysis(hls_synth_res_estimation)
    estimate_layer_resources_hls["total"] = aggregate_dict_keys(estimate_layer_resources_hls)
    with open(report_dir + "/estimate_layer_resources_hls.json", "w") as f:
        json.dump(estimate_layer_resources_hls, f, indent=2)

    if VerificationStepType.NODE_BY_NODE_RTLSIM in cfg._resolve_verification_steps():
        model = model.transform(PrepareRTLSim())
        model = model.transform(SetExecMode("rtlsim"))
        verify_step(model, cfg, "node_by_node_rtlsim", need_parent=True)
    return model


def step_set_fifo_depths(model: ModelWrapper, cfg: DataflowBuildConfig):
    """
    Depending on the auto_fifo_depths setting, do one of the following:
    * if auto_fifo_depths=True:  Run the appropriate auto-sizing transformation
    to attempt to determine the FIFO sizes that provide full throughput.
    May take a long time.
    * if auto_fifo_depths=False:  Assume the folding config file contains FIFO
    sizes as well. Runs the `InsertFIFO` transformation, then
    `ApplyConfig(cfg.folding_config_file)`, and finally `RemoveShallowFIFOs`.
    Coherency with config file node naming is ensured by calling
    `GiveUniqueNodeNames`.
    """

    hw_attrs = [
        "PE",
        "SIMD",
        "parallel_window",
        "ram_style",
        "depth",
        "impl_style",
        "resType",
        "mem_mode",
        "runtime_writeable_weights",
        "inFIFODepths",
        "outFIFODepths",
        "depth_trigger_uram",
        "depth_trigger_bram",
    ]

    # Experimental live FIFO-sizing, overwrites all other FIFO-related behavior
    if cfg.live_fifo_sizing:
        # Create all DWCs and FIFOs normally
        model = model.transform(InsertDWC())
        model = model.transform(
            InsertFIFO(vivado_ram_style=cfg.large_fifo_mem_style, create_shallow_fifos=True)
        )

        # Clean up model
        model = model.transform(SortGraph())
        model = model.transform(GiveUniqueNodeNames())
        model = model.transform(GiveReadableTensorNames())

        # save original folding config before potentially modifying it
        cfg_path = cfg.output_dir + "/report/folding_config_before_lfs.json"
        extract_model_config_to_json(model, cfg_path, hw_attrs)
        model.set_metadata_prop("folding_config_before_lfs", cfg_path)

        # Disable runtime-writable weights, external weights, and dynamic mode,
        # as we don't support additional AXI-lite interfaces next to the FIFOs
        for node in model.graph.node:
            if node.domain.startswith("finn.custom_op.fpgadataflow"):
                node_inst = getCustomOp(node)
                try:
                    if node_inst.get_nodeattr("runtime_writeable_weights") == 1:
                        node_inst.set_nodeattr("runtime_writeable_weights", 0)
                        if node_inst.get_nodeattr("ram_style") == "ultra":
                            node_inst.set_nodeattr("ram_style", "block")
                except AttributeError:
                    pass
                try:
                    if node_inst.get_nodeattr("mem_mode") == "external":
                        node_inst.set_nodeattr("mem_mode", "internal_decoupled")
                except AttributeError:
                    pass
                try:
                    if node_inst.get_nodeattr("dynamic_mode") == 1:
                        node_inst.set_nodeattr("dynamic_mode", 0)
                except AttributeError:
                    pass

        # Specialize FIFOs to HLS back-end instead of default RTL back-end
        for node in model.get_nodes_by_op_type("StreamingFIFO"):
            node_inst = getCustomOp(node)
            node_inst.set_nodeattr("preferred_impl_style", "hls")
        model = model.transform(SpecializeLayers(cfg._resolve_fpga_part()))

        # Fix impl_style attribute
        for node in model.get_nodes_by_op_type("StreamingFIFO_hls"):
            node_inst = getCustomOp(node)
            node_inst.set_nodeattr("impl_style", "virtual")

        # Clean up model
        model = model.transform(SortGraph())
        model = model.transform(GiveUniqueNodeNames())
        model = model.transform(GiveReadableTensorNames())

        return model

    if cfg.auto_fifo_depths:
        if cfg.auto_fifo_strategy == "characterize":
            model = model.transform(InsertDWC())
            model = model.transform(SpecializeLayers(cfg._resolve_fpga_part()))
            model = model.transform(GiveUniqueNodeNames())
            model = model.transform(
                PrepareIP(cfg._resolve_fpga_part(), cfg._resolve_hls_clk_period())
            )
            model = model.transform(HLSSynthIP())
            model = model.transform(PrepareRTLSim())
            model = model.transform(AnnotateCycles())
            period = model.analysis(dataflow_performance)["max_cycles"] + 10
            model = model.transform(DeriveCharacteristic(period))
            model = model.transform(DeriveFIFOSizes())
            model = model.transform(
                InsertFIFO(
                    vivado_ram_style=cfg.large_fifo_mem_style,
                    max_qsrl_depth=256,
                    create_shallow_fifos=True,
                )
            )
            model = model.transform(SpecializeLayers(cfg._resolve_fpga_part()))
            model = model.transform(GiveUniqueNodeNames())
            model = model.transform(GiveReadableTensorNames())
        elif cfg.auto_fifo_strategy == "largefifo_rtlsim":
            if cfg.fifosim_save_waveform:
                report_dir = cfg.output_dir + "/report"
                os.makedirs(report_dir, exist_ok=True)
                model.set_metadata_prop(
                    "rtlsim_trace", os.path.abspath(report_dir) + "/fifosim_trace.wdb"
                )
            model = model.transform(
                InsertAndSetFIFODepths(
                    cfg._resolve_fpga_part(),
                    cfg._resolve_hls_clk_period(),
                    swg_exception=cfg.default_swg_exception,
                    vivado_ram_style=cfg.large_fifo_mem_style,
                    fifosim_input_throttle=cfg.fifosim_input_throttle,
                )
            )
            # InsertAndSetFIFODepths internally removes any shallow FIFOs
            # so no need to call RemoveShallowFIFOs here
        else:
            assert "Unsupported auto_fifo_strategy: " + cfg.auto_fifo_strategy
    else:
        # assume folding cfg json contains FIFO sizes too
        # insert DWCs, FIFOs and run ApplyConfig once more
        model = model.transform(InsertDWC())
        # need to make sure all FIFOs are created so that their depth can be
        # set by ApplyConfig, so create_shallow_fifos=True
        model = model.transform(InsertFIFO(create_shallow_fifos=True))
        model = model.transform(SpecializeLayers(cfg._resolve_fpga_part()))
        model = model.transform(GiveUniqueNodeNames())
        model = model.transform(GiveReadableTensorNames())
        if cfg.folding_config_file is not None:
            model = model.transform(ApplyConfig(cfg.folding_config_file))

    # extract the final configuration and save it as json
    extract_model_config_to_json(model, cfg.output_dir + "/report/final_hw_config.json", hw_attrs)

    # perform FIFO splitting and shallow FIFO removal only after the final config
    # json file has been written. otherwise, since these transforms may add/remove
    # FIFOs, we get name mismatch problems when trying to reuse the final config.
    if cfg.split_large_fifos:
        model = model.transform(SplitLargeFIFOs())
    model = model.transform(RemoveShallowFIFOs())

    # generate a dedicated report about final FIFO sizes
    fifo_info = {}
    fifo_info["fifo_depths"] = {}
    fifo_info["fifo_sizes"] = {}
    total_fifo_size = 0
    for node in model.get_nodes_by_op_type("StreamingFIFO_rtl"):
        node_inst = getCustomOp(node)
        fifo_info["fifo_depths"][node.name] = node_inst.get_nodeattr("depth")
        fifo_info["fifo_sizes"][
            node.name
        ] = node_inst.get_instream_width() * node_inst.get_nodeattr("depth")
        total_fifo_size += fifo_info["fifo_sizes"][node.name]
    fifo_info["total_fifo_size_kB"] = int(total_fifo_size / 8.0 / 1000.0)

    with open(cfg.output_dir + "/report/fifo_sizing.json", "w") as f:
        json.dump(fifo_info, f, indent=2)

    # With this step moved before step_hw_codegen and step_hw_ipgen, the following
    # could be removed, but we keep it for now for backwards compatibility:

    # after FIFOs are ready to go, call PrepareIP and HLSSynthIP again
    # this will only run for the new nodes (e.g. FIFOs and DWCs)
    model = model.transform(PrepareIP(cfg._resolve_fpga_part(), cfg._resolve_hls_clk_period()))
    model = model.transform(HLSSynthIP())
    return model


def step_create_stitched_ip(model: ModelWrapper, cfg: DataflowBuildConfig):
    """Create stitched IP for a graph after all HLS IP blocks have been generated.
    Depends on the DataflowOutputType.STITCHED_IP output product."""

    # introduce tLAST marker, required for instrumentation
    if cfg.enable_instrumentation:
        if cfg.shell_flow_type == ShellFlowType.VITIS_ALVEO:
            raise FINNUserError("Instrumentation is not yet implemented for Alveo/Vitis flow")
        model = model.transform(
            InsertTLastMarker(
                # only insert marker on output (input TLAST is ignored for these use-cases anyway)
                both=False,
                # use ap_axiu instead of qdma_axis
                external=False,
                # static number of iterations (based on what the compiler/folding sets up)
                dynamic=False,
            )
        )
        # give a proper name to the inserted node, important for codegen
        # TODO: deal with multi-I/O accelerators?
        model.graph.node[-1].name = "TLastMarker_0"
        # re-run codegen and HLS IP gen, will affect only the new TLastMarker layer assuming
        # all other IPs have been generated already
        model = model.transform(PrepareIP(cfg._resolve_fpga_part(), cfg._resolve_hls_clk_period()))
        model = model.transform(HLSSynthIP())

    if DataflowOutputType.STITCHED_IP in cfg.generate_outputs:
        stitched_ip_dir = cfg.output_dir + "/stitched_ip"
        model = model.transform(
            CreateStitchedIP(
                cfg._resolve_fpga_part(),
                cfg.synth_clk_period_ns,
                vitis=cfg.stitched_ip_gen_dcp,
                signature=cfg.signature,
            )
        )
        # TODO copy all ip sources into output dir? as zip?
        shutil.copytree(
            model.get_metadata_prop("vivado_stitch_proj"), stitched_ip_dir, dirs_exist_ok=True
        )
        log.info(f"Vivado stitched IP written into {stitched_ip_dir}")
    if VerificationStepType.STITCHED_IP_RTLSIM in cfg._resolve_verification_steps():
        # prepare ip-stitched rtlsim
        verify_model = deepcopy(model)
        verify_model = prepare_for_stitched_ip_rtlsim(verify_model, cfg)

        # Use critical path estimate to set rtlsim liveness threshold
        # TODO: This is a heuristic which usually overestimates the maximum
        #  cycles (by a lot), but can actually also underestimate causing
        #  incorrect detection of timeouts. In these cases, this estimation can
        #  be overwritten by setting LIVENESS_THRESHOLD to a very large value.
        verify_model = verify_model.transform(AnnotateCycles())
        liveness = get_liveness_threshold_cycles()
        perf = verify_model.analysis(dataflow_performance)
        latency = perf["critical_path_cycles"]
        max_iters = max(liveness, int(np.ceil(latency * 1.1 + 20)))
        os.environ["LIVENESS_THRESHOLD"] = str(max_iters)

        if cfg.verify_save_rtlsim_waveforms:
            verify_out_dir = cfg.output_dir + "/verification_output"
            os.makedirs(verify_out_dir, exist_ok=True)
            abspath = os.path.abspath(verify_out_dir)
            verify_model.set_metadata_prop("rtlsim_trace", abspath + "/verify_rtlsim.wdb")
        verify_step(verify_model, cfg, "stitched_ip_rtlsim", need_parent=True)
        os.environ["LIVENESS_THRESHOLD"] = str(liveness)
    return model


def step_measure_rtlsim_performance(model: ModelWrapper, cfg: DataflowBuildConfig):
    """Measure performance + latency of stitched-IP model in rtlsim (xsi).
    Depends on the DataflowOutputType.STITCHED_IP output product.
    """

    if DataflowOutputType.RTLSIM_PERFORMANCE in cfg.generate_outputs:
        assert (
            DataflowOutputType.STITCHED_IP in cfg.generate_outputs
        ), "rtlsim_perf needs stitched IP"
        report_dir = cfg.output_dir + "/report"
        os.makedirs(report_dir, exist_ok=True)
        rtlsim_bs = int(cfg.rtlsim_batch_size)
        orig_rtlsim_trace_depth = get_rtlsim_trace_depth()
        assert rtlsim_bs > 0, "rtlsim batch size must be >0"
        if cfg.verify_save_rtlsim_waveforms:
            # set depth to 3 for layer-by-layer visibility
            os.environ["RTLSIM_TRACE_DEPTH"] = "3"
            model.set_metadata_prop(
                "rtlsim_trace",
                "%s/rtlsim_perf_batch_%d.wdb" % (os.path.abspath(report_dir), rtlsim_bs),
            )

        # Use critical path estimate to set the timeout limit for FIFO sim
        # TODO: This is a heuristic which usually overestimates the maximum
        #  cycles (by a lot), but can actually also underestimate causing
        #  incorrect detection of timeouts. In these cases, this estimation can
        #  be overwritten by setting LIVENESS_THRESHOLD to a very large value.
        model = model.transform(AnnotateCycles())
        liveness = get_liveness_threshold_cycles()
        perf = model.analysis(dataflow_performance)
        latency = perf["critical_path_cycles"]
        max_iters = max(liveness, int(np.ceil(latency * 1.1 + 20)))

        rtlsim_perf_dict = xsi_fifosim(model, rtlsim_bs, max_iters=max_iters)
        # keep keys consistent between the Python and C++-styles
        cycles = rtlsim_perf_dict["cycles"]
        clk_ns = cfg.synth_clk_period_ns
        fclk_mhz = 1 / (clk_ns * 0.001)
        runtime_s = (cycles * clk_ns) * (10**-9)
        rtlsim_perf_dict["runtime[ms]"] = runtime_s * 1000
        rtlsim_perf_dict["throughput[images/s]"] = rtlsim_bs / runtime_s
        rtlsim_perf_dict["fclk[mhz]"] = fclk_mhz
        for key, val in rtlsim_perf_dict.items():
            if "max_count" in key:
                del rtlsim_perf_dict[key]
        # estimate stable-state throughput based on latency+throughput
        if rtlsim_bs == 1:
            rtlsim_perf_dict["stable_throughput[images/s]"] = rtlsim_perf_dict[
                "throughput[images/s]"
            ]
        else:
            total_cycles = rtlsim_perf_dict["cycles"]
            latency_cycles = rtlsim_perf_dict["latency_cycles"]
            stablestate_cycles = total_cycles - latency_cycles
            clk_ns = cfg.synth_clk_period_ns
            fclk_mhz = 1 / (clk_ns * 0.001)
            runtime_s = (stablestate_cycles * clk_ns) * (10**-9)
            rtlsim_perf_dict["stable_throughput[images/s]"] = rtlsim_bs / runtime_s

        with open(report_dir + "/rtlsim_performance.json", "w") as f:
            json.dump(rtlsim_perf_dict, f, indent=2)
        if cfg.verify_save_rtlsim_waveforms:
            # restore original trace depth
            os.environ["RTLSIM_TRACE_DEPTH"] = str(orig_rtlsim_trace_depth)

    return model


def step_make_driver(model: ModelWrapper, cfg: DataflowBuildConfig):
    """Create a driver that can be used to interface the generated accelerator.
    Use DataflowBuildConfig to select PYNQ Python or C++ driver."""

    driver_dir = os.path.join(cfg.output_dir, "driver")
    if DataflowOutputType.PYNQ_DRIVER in cfg.generate_outputs:
        # generate PYNQ driver
        if cfg.enable_instrumentation:
            model = model.transform(
                MakePYNQDriverInstrumentation(
                    cfg._resolve_driver_platform(), cfg.synth_clk_period_ns, cfg.live_fifo_sizing
                )
            )
        else:
            model = model.transform(
                MakePYNQDriverIODMA(cfg._resolve_driver_platform(), cfg.validation_dataset)
            )
        shutil.copytree(model.get_metadata_prop("pynq_driver_dir"), driver_dir, dirs_exist_ok=True)
        log.info("PYNQ Python driver written into " + driver_dir)
    elif DataflowOutputType.CPP_DRIVER in cfg.generate_outputs:
        # generate C++ Driver
        model = model.transform(
            MakeCPPDriver(
                cfg._resolve_driver_platform(),
                version=cfg.cpp_driver_version,
                host_mem=cfg.fpga_memory,
            )
        )
        shutil.copytree(
            model.get_metadata_prop("cpp_driver_dir"),
            driver_dir,
            dirs_exist_ok=True,
            copy_function=shutil.copyfile,
        )

        log.info("C++ driver written into " + driver_dir)
    else:
        log.warning(
            "The step step_make_driver is in the build list but will not be executed"
            + " since no driver is selected in generate_outputs in your build.py file!"
        )
    return model


def step_out_of_context_synthesis(model: ModelWrapper, cfg: DataflowBuildConfig):
    """Run out-of-context synthesis and generate reports.
    Depends on the DataflowOutputType.STITCHED_IP output product."""
    if DataflowOutputType.OOC_SYNTH in cfg.generate_outputs:
        assert DataflowOutputType.STITCHED_IP in cfg.generate_outputs, "OOC needs stitched IP"
        model = model.transform(
            SynthOutOfContext(part=cfg._resolve_fpga_part(), clk_period_ns=cfg.synth_clk_period_ns)
        )
        report_dir = cfg.output_dir + "/report"
        os.makedirs(report_dir, exist_ok=True)
        ooc_res_dict = model.get_metadata_prop("res_total_ooc_synth")
        ooc_res_dict = eval(ooc_res_dict)

        estimate_network_performance = model.analysis(dataflow_performance)
        # add some more metrics to estimated performance
        n_clock_cycles_per_sec = float(ooc_res_dict["fmax_mhz"]) * (10**6)
        est_fps = n_clock_cycles_per_sec / estimate_network_performance["max_cycles"]
        ooc_res_dict["estimated_throughput_fps"] = est_fps
        with open(report_dir + "/ooc_synth_and_timing.json", "w") as f:
            json.dump(ooc_res_dict, f, indent=2)
    return model


def step_vivado_power_estimation(model: ModelWrapper, cfg: DataflowBuildConfig):
    """Run Vivado power estimation on the stitched IP after OOC synthesis."""
    if DataflowOutputType.OOC_SYNTH not in cfg.generate_outputs:
        raise FINNUserError("Vivado power estimation needs OOC synth")

    report_dir = cfg.output_dir + "/report"
    model.transform(
        VivadoPowerEstimation(
            report_dir,
            cfg.synth_clk_period_ns,
            cfg.vivado_power_simulate_activity,
            cfg.vivado_power_simulation_type,
        )
    )
    return model


def step_synthesize_bitfile(model: ModelWrapper, cfg: DataflowBuildConfig):
    """Synthesize a bitfile for the using the specified shell flow, using either
    Vivado or Vitis, to target the specified board."""

    if DataflowOutputType.BITFILE in cfg.generate_outputs:
        bitfile_dir = cfg.output_dir + "/bitfile"
        os.makedirs(bitfile_dir, exist_ok=True)
        report_dir = cfg.output_dir + "/report"
        os.makedirs(report_dir, exist_ok=True)
        partition_model_dir = cfg.output_dir + "/intermediate_models/kernel_partitions"
        if cfg.shell_flow_type == ShellFlowType.VIVADO_ZYNQ:
            model = model.transform(
                ZynqBuild(
                    cfg.board,
                    cfg.synth_clk_period_ns,
                    cfg.enable_hw_debug,
                    cfg.enable_instrumentation,
                    partition_model_dir=partition_model_dir,
                )
            )

            bitfile_path = os.path.join(bitfile_dir, "finn-accel.bit")
            copy(model.get_metadata_prop("bitfile"), bitfile_path)
            copy(model.get_metadata_prop("hw_handoff"), bitfile_dir + "/finn-accel.hwh")
            copy(
                model.get_metadata_prop("vivado_synth_rpt"),
                report_dir + "/post_synth_resources.xml",
            )

            model.set_metadata_prop("bitfile_output", os.path.abspath(bitfile_path))

            post_synth_resources = model.analysis(post_synth_res)
            with open(report_dir + "/post_synth_resources.json", "w") as f:
                json.dump(post_synth_resources, f, indent=2)

            vivado_pynq_proj_dir = model.get_metadata_prop("vivado_pynq_proj")
            timing_rpt = (
                "%s/finn_zynq_link.runs/impl_1/top_wrapper_timing_summary_routed.rpt"
                % vivado_pynq_proj_dir
            )
            copy(timing_rpt, report_dir + "/post_route_timing.rpt")

        elif cfg.shell_flow_type == ShellFlowType.VITIS_ALVEO:
            model = model.transform(
                VitisBuild(
                    cfg._resolve_fpga_part(),
                    cfg.synth_clk_period_ns,
                    cfg._resolve_vitis_platform(),
                    strategy=cfg.vitis_opt_strategy,
                    enable_debug=cfg.enable_hw_debug,
                    floorplan_file=cfg.vitis_floorplan_file,
                    partition_model_dir=partition_model_dir,
                    fpga_memory_type=cfg.fpga_memory,
                )
            )

            bitfile_path = os.path.join(bitfile_dir, "finn-accel.xclbin")
            copy(model.get_metadata_prop("bitfile"), bitfile_path)
            copy(
                model.get_metadata_prop("vivado_synth_rpt"),
                report_dir + "/post_synth_resources.xml",
            )

            model.set_metadata_prop("bitfile_output", os.path.abspath(bitfile_path))

            post_synth_resources = model.analysis(post_synth_res)
            with open(report_dir + "/post_synth_resources.json", "w") as f:
                json.dump(post_synth_resources, f, indent=2)
        else:
            raise Exception("Unrecognized shell_flow_type: " + str(cfg.shell_flow_type))
        log.info(f"Bitfile written into {bitfile_dir}")

    return model


def step_deployment_package(model: ModelWrapper, cfg: DataflowBuildConfig):
    """Create a deployment package including the driver and bitfile."""

    if DataflowOutputType.DEPLOYMENT_PACKAGE in cfg.generate_outputs:
        deploy_dir = cfg.output_dir + "/deploy"
        bitfile_dir = cfg.output_dir + "/bitfile"
        driver_dir = cfg.output_dir + "/driver"
        os.makedirs(deploy_dir, exist_ok=True)
        shutil.copytree(bitfile_dir, deploy_dir + "/bitfile", dirs_exist_ok=True)
        shutil.copytree(
            driver_dir, deploy_dir + "/driver", dirs_exist_ok=True, copy_function=shutil.copyfile
        )
        if DataflowOutputType.CPP_DRIVER in cfg.generate_outputs:
            update_bitfile_path_after_copy(
                os.path.join(deploy_dir, "bitfile", "finn-accel.xclbin"),
                os.path.join(deploy_dir, "driver", "acceleratorconfig.json"),
            )

    return model


#: map step name strings to step functions
build_dataflow_step_lookup = {
    "step_qonnx_to_finn": step_qonnx_to_finn,
    "step_tidy_up": step_tidy_up,
    "step_streamline": step_streamline,
    "step_convert_to_hw": step_convert_to_hw,
    "step_specialize_layers": step_specialize_layers,
    "step_create_dataflow_partition": step_create_dataflow_partition,
    "step_target_fps_parallelization": step_target_fps_parallelization,
    "step_apply_folding_config": step_apply_folding_config,
    "step_minimize_bit_width": step_minimize_bit_width,
    "step_generate_estimate_reports": step_generate_estimate_reports,
    "step_hw_codegen": step_hw_codegen,
    "step_hw_ipgen": step_hw_ipgen,
    "step_set_fifo_depths": step_set_fifo_depths,
    "step_create_stitched_ip": step_create_stitched_ip,
    "step_measure_rtlsim_performance": step_measure_rtlsim_performance,
    "step_make_driver": step_make_driver,
    "step_out_of_context_synthesis": step_out_of_context_synthesis,
    "step_vivado_power_estimation": step_vivado_power_estimation,
    "step_synthesize_bitfile": step_synthesize_bitfile,
    "step_deployment_package": step_deployment_package,
}



# Copyright (C) 2023, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from math import ceil
from qonnx.core.datatype import DataType

from finn.custom_op.fpgadataflow.channelwise_op import ChannelwiseOp
from finn.custom_op.fpgadataflow.hlsbackend import HLSBackend
from finn.util.data_packing import numpy_to_hls_code

# ONNX i/o tensor shape assumptions for channelwise ops:
# input 0 is the input tensor, shape (..., NumChannels)
# input 1 is the channelwise parameter tensor, shape (NumChannels, params_per_channel)
# output 0 is the output tensor, shape (..., NumChannels) - same as input
# the ... here can be any shape (representing groups of vectors)


class ChannelwiseOp_hls(ChannelwiseOp, HLSBackend):
    """Class that corresponds to finn-hls Thresholding_Batch function.
    It can implement a variety of channel-wise parametrized operations,
    including Add, Mul and multi-thresholding.
    """

    def __init__(self, onnx_node, **kwargs):
        super().__init__(onnx_node, **kwargs)

    def get_nodeattr_types(self):
        my_attrs = {}
        my_attrs.update(ChannelwiseOp.get_nodeattr_types(self))
        my_attrs.update(HLSBackend.get_nodeattr_types(self))
        return my_attrs

    def verify_node(self):
        info_messages = []
        # verify that "backend" is set to "fpgadataflow"
        backend_value = self.get_nodeattr("backend")
        if backend_value == "fpgadataflow":
            info_messages.append("Attribute backend is set correctly")
        else:
            info_messages.append('Attribute backend should be set to "fpgadataflow"')

        # verify that all necessary attributes exist
        # TODO collect automatically from get_nodeattr_types
        try:
            self.get_nodeattr("code_gen_dir_cppsim")
            self.get_nodeattr("executable_path")
            self.get_nodeattr("NumChannels")
            self.get_nodeattr("PE")
            self.get_nodeattr("inputDataType")
            self.get_nodeattr("paramDataType")
            self.get_nodeattr("outputDataType")
            info_messages.append("All necessary attributes exist")
        except Exception:
            info_messages.append("""The required Threshold_Batch attributes do not exist.""")

        return info_messages

    def bram_estimation(self):
        """Calculates BRAM cost if resource set to BRAM"""
        style = self.get_nodeattr("ram_style")
        P = self.get_nodeattr("PE")
        idt = self.get_input_datatype()
        A = idt.bitwidth()
        tmem = self.calc_tmem()

        if style == "block" and tmem > 1:
            return int(ceil(A * P / 16)) * int(ceil(tmem / 1024))
        else:
            return 0

    def lut_estimation(self):
        """Calculates LUT cost, taking memory resource type into account"""
        # TODO add in/out FIFO contributions
        style = self.get_nodeattr("ram_style")
        P = self.get_nodeattr("PE")
        idt = self.get_input_datatype()
        A = idt.bitwidth()
        tmem = self.calc_tmem()
        # cost of comparators
        comparator_cost = A * P
        # cost of LUTRAM
        if style == "distributed" and tmem > 1:
            lutram_cost = P * A * int(ceil(tmem / 64))
        else:
            lutram_cost = 0
        # total cost
        return comparator_cost + lutram_cost

    def get_template_param_values(self):
        """Returns the template parameter values according to input, output and weight
        data types."""
        ret = dict()
        inp_hls_str = self.get_input_datatype().get_hls_datatype_str()
        out_hls_str = self.get_output_datatype().get_hls_datatype_str()
        # fill in TSrcI
        ret["TSrcI"] = "Slice<%s>" % inp_hls_str
        # fill in TDstI
        ret["TDstI"] = "Slice<%s>" % out_hls_str

        return ret

    def get_hls_compatible_parameter_tensor(self, orig_param_vector):
        """Convert the original numpy weight matrix orig_weight_matrix into
        a form suitable for passing to the hlslib call:
        * ensure chn % PE == 0
        * interleave rows between PEs
        * reshape into (PE, TMEM) and return
        """
        chn = self.get_nodeattr("NumChannels")
        pe = self.get_nodeattr("PE")
        tmem = chn // pe
        assert chn % pe == 0, "Requirement NumChannels divisable by PE is violated."
        assert (
            orig_param_vector.ndim == 1
        ), """Parameter vector dimension is {}.
        Expected dimension: 1.""".format(
            orig_param_vector.ndim
        )

        # if not self.get_input_datatype().signed():
        #     # ensure all thresholds are nonnegative
        #     assert (orig_param_vector >= 0).all()

        # ensure all thresholds are integer
        assert (orig_param_vector.astype(np.int32) == orig_param_vector).all()
        ret = orig_param_vector

        assert ret.shape[0] == chn, "Cardinality of parameter vector is not as expected (chn)"

        # distribute rows between PEs
        ret = ret.reshape(tmem, pe).transpose()
        assert (
            ret.shape[0] == pe
        ), """First dimension after distribution of the
        rows between PEs is not as expected (pe)"""
        assert (
            ret.shape[1] == tmem
        ), """Second dimension after distribution of the
        rows between PEs is not as expected (tmem)"""

        return ret.reshape(1, pe, tmem)

    def generate_params(self, model, path):
        code_gen_dir = path
        # save thresholds in params.h
        parameters = model.get_initializer(self.onnx_node.input[1])
        parameter_tensor = self.get_hls_compatible_parameter_tensor(parameters)
        pdt = DataType[self.get_nodeattr("paramDataType")]

        parameters_hls_code = numpy_to_hls_code(parameter_tensor, pdt, "parameters", False, True)
        # get input data type
        export_idt = self.get_input_datatype()
        if self.get_input_datatype() == DataType["BIPOLAR"]:
            export_idt = DataType["BINARY"]
        idt_hls = export_idt.get_hls_datatype_str()

        # write parameters into params.h
        f_params = open("{}/params.h".format(code_gen_dir), "w")
        pdt_hls = pdt.get_hls_datatype_str()
        # use binary to export bipolar activations
        export_odt = self.get_output_datatype()
        if self.get_output_datatype() == DataType["BIPOLAR"]:
            export_odt = DataType["BINARY"]
        odt_hls = export_odt.get_hls_datatype_str()
        # get desired function
        func = self.get_nodeattr("Func")
        if func == "cmp_le":
            func_str = "comp::less_equal<%s, %s>" % (idt_hls, pdt_hls)
        elif func == "cmp_ge":
            func_str = "comp::greater_equal<%s, %s>" % (idt_hls, pdt_hls)
        elif func == "add":
            func_str = "comp::add<%s, %s, %s>" % (pdt_hls, idt_hls, odt_hls)
        elif func == "mul":
            func_str = "comp::mul<%s, %s, %s>" % (pdt_hls, idt_hls, odt_hls)
        else:
            raise Exception(
                """Invalid value for attribute Func! Is currently set to: {}
            has to be set to one of the following value
            ("cmp_le", "cmp_ge", "add", "mul")""".format(
                    func
                )
            )
        f_params.write(
            "static ChannelWiseOperation<{},{},{},{},{},{}> threshs \
            = ".format(
                self.calc_tmem(),
                self.get_nodeattr("PE"),
                idt_hls,
                pdt_hls,
                odt_hls,
                func_str,
            )
        )
        f_params.write(parameters_hls_code)
        f_params.close()

    def execute_node(self, context, graph):
        HLSBackend.execute_node(self, context, graph)

    def global_includes(self):
        self.code_gen_dict["$GLOBALS$"] = ['#include "activations.hpp"']
        self.code_gen_dict["$GLOBALS$"] += ['#include "params.h"']

    def read_npy_data(self):
        code_gen_dir = self.get_nodeattr("code_gen_dir_cppsim")
        dtype = self.get_input_datatype()
        elem_bits = dtype.bitwidth()
        packed_bits = self.get_instream_width()
        packed_hls_type = "ap_uint<%d>" % packed_bits
        elem_hls_type = dtype.get_hls_datatype_str()
        npy_type = "float"
        npy_in = "%s/input_0.npy" % code_gen_dir
        self.code_gen_dict["$READNPYDATA$"] = []
        # note: the innermost dim is reversed for the input
        self.code_gen_dict["$READNPYDATA$"].append(
            'npy2apintstream<%s, %s, %d, %s>("%s", in0_V, false);'
            % (
                packed_hls_type,
                elem_hls_type,
                elem_bits,
                npy_type,
                npy_in,
            )
        )

    def defines(self, var):
        numInputVectors = list(self.get_nodeattr("numInputVectors"))
        numReps = numInputVectors[0]
        self.code_gen_dict["$DEFINES$"] = [
            """#define NumChannels1 {}\n#define PE1 {}\n#define numReps {}""".format(
                self.get_nodeattr("NumChannels"),
                self.get_nodeattr("PE"),
                numReps,
            )
        ]

    def docompute(self):
        tmpl_args = self.get_template_param_values()
        # TODO: why put some template parameters into defines and not others?
        # should ImgDim be defined or just filled in here like we do now?
        ishape = self.get_folded_input_shape()
        if len(ishape) == 3:
            spatial_dim = 1
        elif len(ishape) == 5:
            spatial_dim = ishape[1] * ishape[2]
        else:
            raise Exception("""Unexpeted input shape""")
        self.code_gen_dict["$DOCOMPUTE$"] = [
            """Thresholding_Batch<{}, NumChannels1, PE1, {}, {}>
            (in0_V, out0_V, threshs, numReps);""".format(
                spatial_dim,
                tmpl_args["TSrcI"],
                tmpl_args["TDstI"],
            )
        ]

    def dataoutstrm(self):
        code_gen_dir = self.get_nodeattr("code_gen_dir_cppsim")
        dtype = self.get_output_datatype()
        if dtype == DataType["BIPOLAR"]:
            # use binary for bipolar storage
            dtype = DataType["BINARY"]
        elem_bits = dtype.bitwidth()
        packed_bits = self.get_outstream_width()
        packed_hls_type = "ap_uint<%d>" % packed_bits
        elem_hls_type = dtype.get_hls_datatype_str()
        npy_type = "float"
        npy_out = "%s/output_0.npy" % code_gen_dir
        shape = self.get_folded_output_shape()
        shape_cpp_str = str(shape).replace("(", "{").replace(")", "}")

        # note: the innermost dim is not reversed for the output
        self.code_gen_dict["$DATAOUTSTREAM$"] = [
            'apintstream2npy<%s, %s, %d, %s>(out0_V, %s, "%s", false);'
            % (
                packed_hls_type,
                elem_hls_type,
                elem_bits,
                npy_type,
                shape_cpp_str,
                npy_out,
            )
        ]

    def blackboxfunction(self):
        self.code_gen_dict["$BLACKBOXFUNCTION$"] = [
            """void {}(hls::stream<ap_uint<{}>> &in0_V,
                hls::stream<ap_uint<{}>> &out0_V
                )""".format(
                self.onnx_node.name,
                self.get_instream_width(),
                self.get_outstream_width(),
            )
        ]

    def pragmas(self):
        self.code_gen_dict["$PRAGMAS$"] = ["#pragma HLS INTERFACE axis port=in0_V"]
        self.code_gen_dict["$PRAGMAS$"].append("#pragma HLS INTERFACE axis port=out0_V")
        self.code_gen_dict["$PRAGMAS$"].append("#pragma HLS INTERFACE ap_ctrl_none port=return")

        # the channelwise parameter tensor is acc_type [PE][TMEM][N_PARAMS_PER_CHANNEL]
        # partition for parallel access along PE and N_PARAMS_PER_CHANNEL
        # dimensions (dims 1 and 3)
        self.code_gen_dict["$PRAGMAS$"].append(
            ("#pragma HLS ARRAY_PARTITION variable=threshs.parameters " "complete dim=1")
        )
        # set resource type
        ram_style = self.get_nodeattr("ram_style")
        pe = self.get_nodeattr("PE")
        ich = self.get_nodeattr("NumChannels")
        # if PE less than NumChannels, assign cores according to ram_style;
        # otherwise if PE == NumChannels, Vivado HLS will unroll to FFs
        if pe < ich:
            if ram_style == "distributed":
                self.code_gen_dict["$PRAGMAS$"].append(
                    ("#pragma HLS RESOURCE variable=threshs.parameters " "core=ROM_2P_LUTRAM")
                )
            elif ram_style == "block":
                self.code_gen_dict["$PRAGMAS$"].append(
                    ("#pragma HLS RESOURCE variable=threshs.parameters " "core=ROM_2P_BRAM")
                )
            else:
                raise Exception(
                    """Invalid value for attribute ram_style! Is currently set to: {}
                has to be set to one of ("block", "distributed")""".format(
                        ram_style
                    )
                )



# Copyright (C) 2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import math
from qonnx.core.datatype import DataType

from finn.custom_op.fpgadataflow.hwcustomop import HWCustomOp
from finn.util.logging import log


class StreamingFIFO(HWCustomOp):
    def __init__(self, onnx_node, **kwargs):
        super().__init__(onnx_node, **kwargs)

    def get_nodeattr_types(self):
        my_attrs = super().get_nodeattr_types()
        my_attrs.update(
            {
                # FIFO depth
                "depth": ("i", True, 0),
                # folded shape of input/output
                "folded_shape": ("ints", True, []),
                # normal shape of input/output
                "normal_shape": ("ints", True, []),
                # FINN DataTypes for inputs/outputs
                "dataType": ("s", True, ""),
                # FPGA resource type for FIFOs when impl_style is vivado
                # auto -- let Vivado decide
                # block -- use BRAM
                # distributed -- use LUTRAM
                # ultra -- use URAM (on UltraScale+)
                "ram_style": (
                    "s",
                    False,
                    "auto",
                    {"auto", "block", "distributed", "ultra"},
                ),
                # whether depth monitoring is enabled (impl_style=rtl only)
                "depth_monitor": ("i", False, 0),
                # the FIFO does not need its own FIFOs
                "inFIFODepths": ("ints", False, [0]),
                "outFIFODepths": ("ints", False, [0]),
            }
        )

        return my_attrs

    def infer_node_datatype(self, model):
        node = self.onnx_node
        idt = model.get_tensor_datatype(node.input[0])
        if idt != self.get_input_datatype():
            warn_str = "inputDataType changing for %s: %s -> %s " % (
                node.name,
                str(self.get_input_datatype()),
                str(idt),
            )
            log.warning(warn_str)
        self.set_nodeattr("dataType", idt.name)
        # data type stays the same
        model.set_tensor_datatype(node.output[0], idt)

    def get_verilog_top_module_intf_names(self):
        ret = super().get_verilog_top_module_intf_names()
        try:
            is_rtl = self.get_nodeattr("impl_style") == "rtl"
        except AttributeError:
            raise Exception(
                self.onnx_node.name
                + """ is still in hw abstraction format,
                Please run SpecializeLayers() before proceeding."""
            )
        is_depth_monitor = self.get_nodeattr("depth_monitor") == 1
        if is_rtl and is_depth_monitor:
            ret["ap_none"] = ["maxcount"]
        return ret

    def get_normal_input_shape(self, ind=0):
        try:
            depth = self.get_adjusted_depth()
        except AttributeError:
            depth = self.get_nodeattr("depth")
        assert depth >= 1, """Depth is too low"""
        try:
            impl_style = self.get_nodeattr("impl_style") == "rtl"
        except AttributeError:
            impl_style = ""
        if depth > 256 and impl_style == "rtl":
            log.warning("Depth is high, set between 2 and 256 for efficient SRL implementation")
        return self.get_nodeattr("normal_shape")

    def get_normal_output_shape(self, ind=0):
        return self.get_normal_input_shape()

    def get_folded_input_shape(self, ind=0):
        return self.get_nodeattr("folded_shape")

    def get_folded_output_shape(self, ind=0):
        return self.get_nodeattr("folded_shape")

    def get_instream_width(self, ind=0):
        dtype = DataType[self.get_nodeattr("dataType")]
        folded_shape = self.get_nodeattr("folded_shape")
        in_width = folded_shape[-1] * dtype.bitwidth()
        return in_width

    def get_outstream_width(self, ind=0):
        dtype = DataType[self.get_nodeattr("dataType")]
        folded_shape = self.get_nodeattr("folded_shape")
        in_width = folded_shape[-1] * dtype.bitwidth()
        return in_width

    def get_input_datatype(self, ind=0):
        return DataType[self.get_nodeattr("dataType")]

    def get_output_datatype(self, ind=0):
        return DataType[self.get_nodeattr("dataType")]

    def execute_node(self, context, graph):
        node = self.onnx_node
        context[node.output[0]] = context[node.input[0]]

    def bram_estimation(self):
        """Calculates resource estimation for BRAM"""
        try:
            impl = self.get_nodeattr("impl_style") == "rtl"
        except AttributeError:
            raise Exception(
                self.onnx_node.name
                + """ is still in hw abstraction format,
                Please run SpecializeLayers() before proceeding."""
            )

        ram_type = self.get_nodeattr("ram_style")
        try:
            depth = self.get_adjusted_depth()
        except AttributeError:
            depth = self.get_nodeattr("depth")
        W = self.get_instream_width()

        if impl == "rtl" or (impl == "vivado" and ram_type != "block"):
            # Non-BRAM based implementation
            return 0

        if W == 1:
            return math.ceil(depth / 16384)
        elif W == 2:
            return math.ceil(depth / 8192)
        elif W <= 4:
            return (math.ceil(depth / 4096)) * (math.ceil(W / 4))
        elif W <= 9:
            return (math.ceil(depth / 2048)) * (math.ceil(W / 9))
        elif W <= 18 or depth > 512:
            return (math.ceil(depth / 1024)) * (math.ceil(W / 18))
        else:
            return (math.ceil(depth / 512)) * (math.ceil(W / 36))

    def uram_estimation(self):
        """Calculates resource estimation for URAM"""

        try:
            impl = self.get_nodeattr("impl_style") == "rtl"
        except AttributeError:
            raise Exception(
                self.onnx_node.name
                + """ is still in hw abstraction format,
                Please run SpecializeLayers() before proceeding."""
            )
        ram_type = self.get_nodeattr("ram_style")
        try:
            depth = self.get_adjusted_depth()
        except AttributeError:
            depth = self.get_nodeattr("depth")
        W = self.get_instream_width()

        if impl == "rtl" or (impl == "vivado" and ram_type != "ultra"):
            # Non-BRAM based implementation
            return 0
        else:
            return (math.ceil(depth / 4096)) * (math.ceil(W / 72))

    def bram_efficiency_estimation(self):
        try:
            depth = self.get_adjusted_depth()
        except AttributeError:
            depth = self.get_nodeattr("depth")
        W = self.get_instream_width()
        bram16_est = self.bram_estimation()
        if bram16_est == 0:
            return 1
        wbits = W * depth
        bram16_est_capacity = bram16_est * 36 * 512
        return wbits / bram16_est_capacity

    def lut_estimation(self):
        """Calculates resource estimations for LUTs"""
        try:
            impl = self.get_nodeattr("impl_style") == "rtl"
        except AttributeError:
            raise Exception(
                self.onnx_node.name
                + """ is still in hw abstraction format,
                Please run SpecializeLayers() before proceeding."""
            )
        ram_type = self.get_nodeattr("ram_style")
        try:
            depth = self.get_adjusted_depth()
        except AttributeError:
            depth = self.get_nodeattr("depth")
        W = self.get_instream_width()

        address_luts = 2 * math.ceil(math.log(depth, 2))

        if impl == "rtl" or (impl == "vivado" and ram_type == "distributed"):
            ram_luts = (math.ceil(depth / 32)) * (math.ceil(W / 2))
        else:
            ram_luts = 0

        return int(address_luts + ram_luts)



# Copyright (C) 2023, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np

from finn.custom_op.fpgadataflow.hlsbackend import HLSBackend
from finn.custom_op.fpgadataflow.streamingeltwise import StreamingEltwise


class StreamingEltwise_hls(StreamingEltwise, HLSBackend):
    """Class that corresponds to finn-hlslib StreamingEltwise function."""

    def __init__(self, onnx_node, **kwargs):
        super().__init__(onnx_node, **kwargs)

    def get_nodeattr_types(self):
        my_attrs = {}
        my_attrs.update(StreamingEltwise.get_nodeattr_types(self))
        my_attrs.update(HLSBackend.get_nodeattr_types(self))
        return my_attrs

    def verify_node(self):
        info_messages = []
        # verify that "backend" is set to "fpgadataflow"
        backend_value = self.get_nodeattr("backend")
        if backend_value == "fpgadataflow":
            info_messages.append("Attribute backend is set correctly")
        else:
            info_messages.append('Attribute backend should be set to "fpgadataflow"')

        # verify that all necessary attributes exist
        try:
            self.get_nodeattr("code_gen_dir_cppsim")
            self.get_nodeattr("executable_path")
            self.get_nodeattr("NumChannels")
            self.get_nodeattr("PE")
            self.get_nodeattr("inputDataType0")
            self.get_nodeattr("inputDataType1")
            self.get_nodeattr("eltwiseOp")
            info_messages.append("All necessary attributes exist")
        except Exception:
            info_messages.append("""The required StreamingEltwise attributes do not exist.""")

        return info_messages

    def execute_node(self, context, graph):
        HLSBackend.execute_node(self, context, graph)

    def global_includes(self):
        self.code_gen_dict["$GLOBALS$"] = [
            '#include "eltwise.hpp"',
            '#include "interpret.hpp"',
        ]

        self.code_gen_dict["$GLOBALS$"].extend(
            [
                "template<typename TI1, typename TI2, typename TO>",
                "struct absdiff {",
                "TO operator()(TI1 const &a, TI2 const &b) const {",
                "#pragma HLS inline",
                "return  a>b? a-b : b-a;",
                "}",
                "};",
                "template<typename TI1, typename TI2, typename TO>",
                "struct sub {",
                "TO operator()(TI1 const &a, TI2 const &b) const {",
                "#pragma HLS inline",
                "return  a-b;",
                "}",
                "};",
                "template<typename TI1, typename TI2, typename TO>",
                "struct add {",
                "TO operator()(TI1 const &a, TI2 const &b) const {",
                "#pragma HLS inline",
                "return  a+b;",
                "}",
                "};",
            ]
        )

    def defines(self, var):
        self.code_gen_dict["$DEFINES$"] = []

    def strm_decl(self):
        self.code_gen_dict["$STREAMDECLARATIONS$"] = []
        self.code_gen_dict["$STREAMDECLARATIONS$"].append(
            'hls::stream<ap_uint<{}>> in0_V ("in0_V");'.format(self.get_instream_width(0))
        )
        self.code_gen_dict["$STREAMDECLARATIONS$"].append(
            'hls::stream<ap_uint<{}>> in1_V ("in1_V");'.format(self.get_instream_width(1))
        )
        self.code_gen_dict["$STREAMDECLARATIONS$"].append(
            'hls::stream<ap_uint<{}>> out0_V ("out0_V");'.format(self.get_outstream_width())
        )

    def docompute(self):
        op = self.get_nodeattr("eltwiseOp")
        idt0 = self.get_input_datatype(0)
        idt1 = self.get_input_datatype(1)
        odt = self.get_output_datatype()
        elem_hls_type_0 = idt0.get_hls_datatype_str()
        elem_hls_type_1 = idt1.get_hls_datatype_str()
        out_hls_type = odt.get_hls_datatype_str()
        slice_in0 = "Slice<%s>" % elem_hls_type_0
        slice_in1 = "Slice<%s>" % elem_hls_type_1
        slice_out = "Slice<%s>" % out_hls_type
        eltwise_op_str = self.get_eltwise_op_lambda()
        "%sEltwiseFunction<%s, %s, %s>()" % (
            op,
            elem_hls_type_0,
            elem_hls_type_1,
            out_hls_type,
        )
        self.code_gen_dict["$DOCOMPUTE$"] = [
            """{}<{}, {}, {}, {}, {}, {}>(in0_V, in1_V, out0_V, {});""".format(
                "StreamingEltwise",
                self.get_nodeattr("NumChannels"),
                self.get_nodeattr("PE"),
                int(np.prod(self.get_folded_output_shape()[:-2])),
                slice_in0,
                slice_in1,
                slice_out,
                eltwise_op_str,
            )
        ]

    def blackboxfunction(self):
        self.code_gen_dict["$BLACKBOXFUNCTION$"] = [
            """void {}(hls::stream<ap_uint<{}>> &in0_V, hls::stream<ap_uint<{}>> &in1_V,
                hls::stream<ap_uint<{}>> &out0_V)""".format(
                self.onnx_node.name,
                self.get_nodeattr("PE") * self.get_input_datatype(0).bitwidth(),
                self.get_nodeattr("PE") * self.get_input_datatype(1).bitwidth(),
                self.get_nodeattr("PE") * self.get_output_datatype().bitwidth(),
            )
        ]

    def pragmas(self):
        self.code_gen_dict["$PRAGMAS$"] = ["#pragma HLS INTERFACE axis port=in0_V"]
        self.code_gen_dict["$PRAGMAS$"].append("#pragma HLS INTERFACE axis port=in1_V")
        self.code_gen_dict["$PRAGMAS$"].append("#pragma HLS INTERFACE axis port=out0_V")
        self.code_gen_dict["$PRAGMAS$"].append("#pragma HLS INTERFACE ap_ctrl_none port=return")



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

try:
    import finn_xsi.adapter as finnxsi
except ModuleNotFoundError:
    finnxsi = None

import numpy as np
import os
from qonnx.custom_op.registry import getCustomOp
from subprocess import CalledProcessError

from finn.util.basic import (
    get_liveness_threshold_cycles,
    get_vivado_root,
    launch_process_helper,
    make_build_dir,
)
from finn.util.data_packing import npy_to_rtlsim_input, rtlsim_output_to_npy
from finn.util.exception import FINNError


def prep_rtlsim_io_dict(model, execution_context):
    # extract i/o info to prepare io_dict
    io_dict = {"inputs": {}, "outputs": {}}
    if_dict = eval(model.get_metadata_prop("vivado_stitch_ifnames"))
    # go over and prepare inputs
    for i, i_vi in enumerate(model.graph.input):
        i_name = i_vi.name
        i_tensor = execution_context[i_name]
        i_dt = model.get_tensor_datatype(i_name)
        first_node_onnx = model.find_consumer(i_name)
        first_node = getCustomOp(first_node_onnx)
        node_inp_ind = list(first_node_onnx.input).index(i_name)
        if node_inp_ind == 0:
            # default node input (input 0)
            i_stream_w = first_node.get_instream_width()
            i_folded_shape = first_node.get_folded_input_shape()
        else:
            # not input 0; node must support specifying inp index
            # for these functions
            i_stream_w = first_node.get_instream_width(node_inp_ind)
            i_folded_shape = first_node.get_folded_input_shape(node_inp_ind)
        batchsize = i_tensor.shape[0]
        # override batch size for input
        i_folded_shape = list(i_folded_shape)
        i_folded_shape[0] = batchsize
        i_folded_shape = tuple(i_folded_shape)
        # TODO any other layout transformations need to happen here!
        i_tensor = i_tensor.reshape(i_folded_shape)
        # pack input for rtlsim
        packed_input = npy_to_rtlsim_input(i_tensor, i_dt, i_stream_w)
        # add to io_dict
        if_name = if_dict["s_axis"][i][0]
        io_dict["inputs"][if_name] = packed_input
    # go over outputs to determine how many values will be produced
    num_out_values = {}
    o_tensor_info = []
    for o, o_vi in enumerate(model.graph.output):
        # output in io_dict just needs an empty list
        if_name = if_dict["m_axis"][o][0]
        io_dict["outputs"][if_name] = []
        # extract output shape
        o_name = o_vi.name
        o_shape = model.get_tensor_shape(o_name)
        o_dt = model.get_tensor_datatype(o_name)
        last_node = getCustomOp(model.find_producer(o_name))
        o_folded_shape = last_node.get_folded_output_shape()
        # override batch size from actual input
        o_shape = list(o_shape)
        if o_shape[0] != batchsize:
            o_shape[0] = batchsize
            num_out_values[if_name] = batchsize * last_node.get_number_output_values()
        else:
            num_out_values[if_name] = last_node.get_number_output_values()
        o_shape = tuple(o_shape)
        o_folded_shape = list(o_folded_shape)
        if o_folded_shape[0] != batchsize:
            o_folded_shape[0] = batchsize
        o_folded_shape = tuple(o_folded_shape)
        o_stream_w = last_node.get_outstream_width()
        o_tensor_info.append((o_stream_w, o_dt, o_folded_shape, o_shape))
    if len(num_out_values.keys()) == 1:
        num_out_values = num_out_values[if_name]
    return io_dict, if_dict, num_out_values, o_tensor_info, batchsize


def file_to_basename(x):
    return os.path.basename(os.path.realpath(x))


def rtlsim_exec_cppxsi(
    model,
    execution_context,
    dummy_data_mode=False,
    timeout_cycles=None,
    throttle_cycles=0,
):
    """Use XSI C++ rtl simulation to execute given model with stitched IP.
    The dummy_data_mode flag controls whether the simulation is driven by
    dummy data or real data. The execution_context parameter must be formatted
    according to whether dummy or real data is used.
    Example with dummy_data = True:
        execution_context = {
            "inputs" : {"<name_of_input_stream>" : <number_of_transactions>},
            "outputs" : {"<name_of_output_stream>" : <number_of_transactions>},
        }
    Example with dummy_data = False:
        execution_context = {
            "<tensor_name>" : <np.ndarray>
        }

    If timeout_cycles is not None, the default value from get_liveness_threshold_cycles
    will be used.
    throttle_cycles will be used to pause the input stream every time an input frame is finished.
    """
    # TODO: support running functional rtlsim with real I/O data
    # TODO: support running with multiple inputs/outputs
    if timeout_cycles is None:
        timeout_cycles = get_liveness_threshold_cycles()

    assert dummy_data_mode, "Only dummy_data_mode=True is supported for now"

    # ensure stitched ip project already exists
    assert os.path.isfile(
        model.get_metadata_prop("wrapper_filename")
    ), """The
    file name from metadata property "wrapper_filename" doesn't exist."""
    assert os.path.isdir(
        model.get_metadata_prop("vivado_stitch_proj")
    ), """The
    directory from metadata property "vivado_stitch_proj" doesn't exist"""
    trace_file = model.get_metadata_prop("rtlsim_trace")
    if not dummy_data_mode:
        # ignore last value which would be batchsize
        io_dict, if_dict, num_out_values, o_tensor_info = prep_rtlsim_io_dict(
            model, execution_context
        )[:-1]

    # prepare rtlsim compiled object (unless it already exists)
    rtlsim_so = model.get_metadata_prop("rtlsim_so")
    top_module_file_name = file_to_basename(model.get_metadata_prop("wrapper_filename"))
    top_module_name = top_module_file_name.strip(".v")
    if (rtlsim_so is None) or (not os.path.isfile(rtlsim_so)):
        vivado_stitch_proj_dir = model.get_metadata_prop("vivado_stitch_proj")
        with open(vivado_stitch_proj_dir + "/all_verilog_srcs.txt", "r") as f:
            all_verilog_srcs = f.read().split()
        single_src_dir = make_build_dir("rtlsim_" + top_module_name + "_")
        debug = not (trace_file is None or trace_file == "")
        rtlsim_so = finnxsi.compile_sim_obj(
            top_module_name, all_verilog_srcs, single_src_dir, debug=debug
        )
        # save generated lib filename in attribute
        model.set_metadata_prop("rtlsim_so", rtlsim_so[0] + "/" + rtlsim_so[1])
        sim_base, sim_rel = rtlsim_so
        # pass in correct tracefile from attribute
        if trace_file == "default":
            trace_file = top_module_file_name + ".wdb"
    else:
        sim_base, sim_rel = rtlsim_so.split("xsim.dir")
        sim_rel = "xsim.dir" + sim_rel
    # prepare the C++ sim driver template
    finnxsi_dir = os.environ["FINN_XSI"]
    fifosim_config_fname = finnxsi_dir + "/rtlsim_config.hpp.template"
    with open(fifosim_config_fname, "r") as f:
        fifsom_config_template = f.read()

    instream_iters = []
    outstream_iters = []
    for top_inp in model.graph.input:
        iname = top_inp.name
        first_node = model.find_consumer(iname)
        assert first_node is not None, "Failed to find consumer for " + iname
        fnode_inst = getCustomOp(first_node)
        top_ind = list(first_node.input).index(iname)
        ishape_folded = fnode_inst.get_folded_input_shape(ind=top_ind)
        instream_iters.append(np.prod(ishape_folded[:-1]))
    for top_out in model.graph.output:
        oname = top_out.name
        last_node = model.find_producer(oname)
        assert last_node is not None, "Failed to find producer for " + oname
        lnode_inst = getCustomOp(last_node)
        top_ind = list(last_node.output).index(oname)
        oshape_folded = lnode_inst.get_folded_output_shape(ind=top_ind)
        outstream_iters.append(np.prod(oshape_folded[:-1]))

    # retrieve the number of inputs from execution_context
    n_inferences = execution_context[model.graph.input[0].name]
    ifnames = model.get_metadata_prop("vivado_stitch_ifnames")
    assert not (
        ifnames is None
    ), "Couldn't find stitched-IP interface names, did you run IP stitching first?"
    ifnames = eval(ifnames)
    if "aximm" in ifnames.keys() and ifnames["aximm"] != []:
        assert (
            False
        ), f"cppxsi sim doesn't know how to handle full AXI MM interfaces: {ifnames['aximm']}"
    instream_names = [x[0] for x in ifnames["s_axis"]]
    outstream_names = [x[0] for x in ifnames["m_axis"]]
    instream_descrs = [
        (instream_names[i], instream_iters[i], instream_iters[i] + throttle_cycles)
        for i in range(len(instream_names))
    ]
    instream_descrs_str = str(instream_descrs).replace("[", "").replace("]", "")
    instream_descrs_str = instream_descrs_str.replace("(", "{").replace(")", "}")
    instream_descrs_str = instream_descrs_str.replace("'", '"')

    outstream_descrs = [
        (outstream_names[i], outstream_iters[i], outstream_iters[i])
        for i in range(len(outstream_names))
    ]
    outstream_descrs_str = str(outstream_descrs).replace("[", "").replace("]", "")
    outstream_descrs_str = outstream_descrs_str.replace("(", "{").replace(")", "}")
    outstream_descrs_str = outstream_descrs_str.replace("'", '"')

    # fill in the template arguments for sim config
    template_dict = {
        # number of inferences
        "N_INFERENCES": n_inferences,
        # max number of cycles to wait for output activity before timeout
        "TIMEOUT_CYCLES": timeout_cycles,
        # name of the top-level HDL module
        "TOP_MODULE_NAME": top_module_name,
        # top-level AXI stream descriptors
        "ISTREAM_DESC": instream_descrs_str,
        "OSTREAM_DESC": outstream_descrs_str,
        # control tracing and trace filename
        "TRACE_FILE": "nullptr" if trace_file is None else f'"{trace_file}"',
        # sim kernel .so to use (depends on Vivado version)
        "SIMKERNEL_SO": finnxsi.get_simkernel_so(),
        # log file for xsi (not the sim driver)
        "XSIM_LOG_FILE": '"xsi.log"',
    }
    for key, val in template_dict.items():
        fifsom_config_template = fifsom_config_template.replace(f"@{key}@", str(val))
    with open(sim_base + "/rtlsim_config.hpp", "w") as f:
        f.write(fifsom_config_template)

    vivado_incl_dir = get_vivado_root() + "/data/xsim/include"
    # launch g++ to compile the rtlsim executable
    build_cmd = [
        "g++",
        f"-I{finnxsi_dir}",
        f"-I{vivado_incl_dir}",
        f"-I{sim_base}",
        "-std=c++17",
        "-O3",
        "-o",
        "rtlsim_xsi",
        f"{finnxsi_dir}/rtlsim_xsi.cpp",
        f"{finnxsi_dir}/xsi_finn.cpp",
        "-ldl",
        "-lrt",
    ]
    # write compilation command to a file for easy re-running/debugging
    with open(sim_base + "/compile_rtlsim.sh", "w") as f:
        f.write(" ".join(build_cmd))
    try:
        launch_process_helper(build_cmd, cwd=sim_base, print_stdout=False)
    except CalledProcessError:
        raise FINNError("Failed to compile rtlsim executable")
    if not os.path.isfile(sim_base + "/rtlsim_xsi"):
        raise FINNError("Failed to compile rtlsim executable")

    # launch the rtlsim executable
    runsim_cmd = ["bash", "run_rtlsim.sh"]
    with open(sim_base + "/run_rtlsim.sh", "w") as f:
        f.write("./rtlsim_xsi > rtlsim_xsi_log.txt")
    launch_process_helper(runsim_cmd, cwd=sim_base)

    # parse results file and return dict
    results_filename = sim_base + "/results.txt"
    with open(results_filename, "r") as f:
        results = f.read().strip().split("\n")
    ret_dict = {}
    for result_line in results:
        key, val = result_line.split("\t")
        ret_dict[key] = int(val)
    if "TIMEOUT" in ret_dict.keys():
        assert ret_dict["TIMEOUT"] == 0, f"XSI C++ simulation timed out, see {results_filename}"
    return ret_dict


def rtlsim_exec_finnxsi(model, execution_context, pre_hook=None, post_hook=None):
    """Use finnxsi to execute given model with stitched IP. The execution
    context contains the input values. Hook functions can be optionally
    specified to observe/alter the state of the circuit
    - pre_hook : hook function to be called before sim start (after reset)
    - post_hook : hook function to be called after sim end
    """
    # ensure stitched ip project already exists
    assert os.path.isfile(
        model.get_metadata_prop("wrapper_filename")
    ), """The
    file name from metadata property "wrapper_filename" doesn't exist."""
    assert os.path.isdir(
        model.get_metadata_prop("vivado_stitch_proj")
    ), """The
    directory from metadata property "vivado_stitch_proj" doesn't exist"""
    trace_file = model.get_metadata_prop("rtlsim_trace")
    io_dict, if_dict, num_out_values, o_tensor_info, batchsize = prep_rtlsim_io_dict(
        model, execution_context
    )

    # prepare rtlsim model
    rtlsim_so = model.get_metadata_prop("rtlsim_so")
    if (rtlsim_so is None) or (not os.path.isfile(rtlsim_so)):
        vivado_stitch_proj_dir = model.get_metadata_prop("vivado_stitch_proj")
        with open(vivado_stitch_proj_dir + "/all_verilog_srcs.txt", "r") as f:
            all_verilog_srcs = f.read().split()
        top_module_file_name = file_to_basename(model.get_metadata_prop("wrapper_filename"))
        top_module_name = top_module_file_name.strip(".v")
        single_src_dir = make_build_dir("rtlsim_" + top_module_name + "_")
        debug = not (trace_file is None or trace_file == "")
        rtlsim_so = finnxsi.compile_sim_obj(
            top_module_name, all_verilog_srcs, single_src_dir, debug=debug
        )
        # save generated lib filename in attribute
        model.set_metadata_prop("rtlsim_so", rtlsim_so[0] + "/" + rtlsim_so[1])
        sim_base, sim_rel = rtlsim_so
        # pass in correct tracefile from attribute
        if trace_file == "default":
            trace_file = top_module_file_name + ".wdb"
        sim = finnxsi.load_sim_obj(sim_base, sim_rel, trace_file)
    else:
        sim_base, sim_rel = rtlsim_so.split("xsim.dir")
        sim_rel = "xsim.dir" + sim_rel
        sim = finnxsi.load_sim_obj(sim_base, sim_rel, trace_file)

    # reset and call rtlsim, including any pre/post hooks
    finnxsi.reset_rtlsim(sim)
    if pre_hook is not None:
        pre_hook(sim)
    n_cycles = finnxsi.rtlsim_multi_io(
        sim,
        io_dict,
        num_out_values,
        sname="",
        liveness_threshold=get_liveness_threshold_cycles() * batchsize,
    )
    if post_hook is not None:
        post_hook(sim)
    # important to call close_rtlsim for finnxsi to flush traces and stop
    finnxsi.close_rtlsim(sim)

    # unpack outputs and put back into execution context
    for o, o_vi in enumerate(model.graph.output):
        o_name = o_vi.name
        if_name = if_dict["m_axis"][o][0]
        o_stream_w, o_dt, o_folded_shape, o_shape = o_tensor_info[o]
        packed_output = io_dict["outputs"][if_name]
        o_folded_tensor = rtlsim_output_to_npy(
            packed_output, None, o_dt, o_folded_shape, o_stream_w, o_dt.bitwidth()
        )
        execution_context[o_name] = o_folded_tensor.reshape(o_shape)

    model.set_metadata_prop("cycles_rtlsim", str(n_cycles))


def rtlsim_exec(model, execution_context, pre_hook=None, post_hook=None):
    """Use XSI to execute given model with stitched IP. The execution
    context contains the input values. Hook functions can be optionally
    specified to observe/alter the state of the circuit, receiving the
    sim object as their first argument:
    - pre_hook : hook function to be called before sim start (after reset)
    - post_hook : hook function to be called after sim end
    """
    rtlsim_exec_finnxsi(model, execution_context, pre_hook, post_hook)



# Copyright (c) 2020-2022, Xilinx, Inc.
# Copyright (C) 2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from finn.custom_op.fpgadataflow.hlsbackend import HLSBackend
from finn.custom_op.fpgadataflow.hwcustomop import HWCustomOp


class TLastMarker_hls(HWCustomOp, HLSBackend):
    """Node that adds/removes AXI stream TLAST signals where needed. Its behavior
    is transparent in node-by-node execution, only visible in IP-stitched rtlsim or
    actual hardware.
    This node  may be needed at the end of the network to signal a DMA write
    (needed by the FINN PYNQ shell) or at the beginning to remove the end-of-burst
    from DMA read."""

    def __init__(self, onnx_node, **kwargs):
        super().__init__(onnx_node, **kwargs)

    def get_nodeattr_types(self):
        my_attrs = {
            # number of (static) iterations until TLAST=1 is generated for Direction=out
            "NumIters": ("i", True, 0),
            # whether static or dynamic (from AXI lite) number of iterations are used
            "DynIters": ("i", False, 1),
            # direction: whether to insert or remove TLAST
            "Direction": ("s", False, "out", {"out", "in"}),
            # width of input-output data streams, in bits
            "StreamWidth": ("i", True, 0),
            # width of individual element in stream, in bits
            "ElemWidth": ("i", True, 0),
            # Protocol: external or internal
            # Vitis docs recommend using qdma_axis for external, ap_axiu for internal
            "Protocol": ("s", False, "external", {"external", "internal"}),
        }
        my_attrs.update(HWCustomOp.get_nodeattr_types(self))
        my_attrs.update(HLSBackend.get_nodeattr_types(self))
        return my_attrs

    def execute_node(self, context, graph):
        # TLastMarker's behavior is only visible when doing
        # rtlsim with stitched IP, since it marks the end
        # of the current image/input sample. when executing
        # inside FINN as a single node, this is not visible.
        # so here we simply return the input as output
        i_name = self.onnx_node.input[0]
        o_name = self.onnx_node.output[0]
        i_tensor = context[i_name]
        context[o_name] = i_tensor

    def make_shape_compatible_op(self, model):
        # not supported for shape inference
        pass

    def infer_node_datatype(self, model):
        # not supported for datatype inference
        pass

    def global_includes(self):
        self.code_gen_dict["$GLOBALS$"] = ['#include "ap_axi_sdata.h"']

    def defines(self, var):
        stream_width = self.get_nodeattr("StreamWidth")
        direction = self.get_nodeattr("Direction")
        protocol = self.get_nodeattr("Protocol")
        # output stream must have TLAST, so we use this stream data type:
        # qdma_axis<stream_data_width,0,0,0 >
        if direction == "out":
            if protocol == "external":
                out_stream_dtype = "qdma_axis<%d,0,0,0>" % stream_width
            elif protocol == "internal":
                out_stream_dtype = "ap_axiu<%d,0,0,0>" % stream_width
            else:
                raise Exception("Unrecognized Protocol in TLastMarker")
            in_stream_dtype = "ap_uint<%d>" % stream_width
        elif direction == "in":
            out_stream_dtype = "ap_uint<%d>" % stream_width
            if protocol == "external":
                in_stream_dtype = "qdma_axis<%d,0,0,0>" % stream_width
            elif protocol == "internal":
                in_stream_dtype = "ap_axiu<%d,0,0,0>" % stream_width
            else:
                raise Exception("Unrecognized Protocol in TLastMarker")
        else:
            raise Exception("Unrecognized Direction in TLastMarker")

        self.code_gen_dict["$DEFINES$"] = [
            "#define StreamWidth %d" % stream_width,
            "#define OutDType %s" % out_stream_dtype,
            "#define InDType %s" % in_stream_dtype,
            "#define NumItersPerImg %d" % self.get_nodeattr("NumIters"),
        ]

    def read_npy_data(self):
        self.code_gen_dict["$READNPYDATA$"] = []

    def docompute(self):
        dyn_iters = self.get_nodeattr("DynIters")
        direction = self.get_nodeattr("Direction")
        use_qdma_axis = self.get_nodeattr("Protocol") == "external"
        if direction == "in":
            # read from input and just pass data along; ignore tlast
            # no dyn iters on input, it doesnt make sense
            self.code_gen_dict["$DOCOMPUTE$"] = [
                "for(unsigned int i=0; i<NumItersPerImg; i++) {",
                "#pragma HLS PIPELINE II=1",
                "out0_V.write(in0_V.read().get_data());"
                if use_qdma_axis
                else "out0_V.write(in0_V.read().data);}",
            ]

        elif dyn_iters == 1:
            # output, with dynamic iteration counts
            self.code_gen_dict["$DOCOMPUTE$"] = [
                "unsigned int n = 1;",
                "OutDType t;",
                "t.set_keep(-1);" if use_qdma_axis else "t.keep = -1;",
                "io_section: { // start of cycle accurate region",
                "#pragma HLS protocol fixed",
                "// do a first read from stream before we decide on numIters",
                "// giving software a chance to set up the numIters prior to startup",
                "t.set_data(in0_V.read());" if use_qdma_axis else "t.data = in0_V.read();",
                "n = (numIters == 0 ? NumItersPerImg : numIters);",
                "t.set_last(n==1);" if use_qdma_axis else "t.last = (n==1);",
                "out0_V.write(t);",
                "} // end of cycle accurate region",
                "// do one less iteration than spec since we already did one",
                "for(unsigned int i=1; i<n; i++) {",
                "#pragma HLS PIPELINE II=1",
                "t.set_data(in0_V.read());" if use_qdma_axis else "t.data = in0_V.read();",
                "t.set_last(i==(n-1));" if use_qdma_axis else "t.last = (i==(n-1));",
                "out0_V.write(t);",
                "}",
            ]

        else:
            # output, with static iteration counts
            self.code_gen_dict["$DOCOMPUTE$"] = [
                "unsigned int n = 1;",
                "OutDType t;",
                "t.set_keep(-1);" if use_qdma_axis else "t.keep = -1;",
                "for(unsigned int i=0; i<NumItersPerImg; i++) {",
                "#pragma HLS PIPELINE II=1",
                "t.set_data(in0_V.read());" if use_qdma_axis else "t.data = in0_V.read();",
                "t.set_last(i==(NumItersPerImg-1));"
                if use_qdma_axis
                else "t.last = (i==(NumItersPerImg-1));",
                "out0_V.write(t);",
                "}",
            ]

    def dataoutstrm(self):
        self.code_gen_dict["$DATAOUTSTREAM$"] = []

    def blackboxfunction(self):
        dyn_iters = self.get_nodeattr("DynIters")

        if dyn_iters == 1:
            self.code_gen_dict["$BLACKBOXFUNCTION$"] = [
                """void %s(hls::stream<InDType> &in0_V,
                    hls::stream<OutDType> &out0_V, unsigned int numIters)"""
                % self.onnx_node.name
            ]
        else:
            self.code_gen_dict["$BLACKBOXFUNCTION$"] = [
                """void %s(hls::stream<InDType> &in0_V,
                hls::stream<OutDType> &out0_V)"""
                % self.onnx_node.name
            ]

    def pragmas(self):
        self.code_gen_dict["$PRAGMAS$"] = ["#pragma HLS INTERFACE axis port=in0_V"]
        self.code_gen_dict["$PRAGMAS$"].append("#pragma HLS INTERFACE axis port=out0_V")

        dyn_iters = self.get_nodeattr("DynIters")
        if dyn_iters == 1:
            self.code_gen_dict["$PRAGMAS$"].append(
                "#pragma HLS INTERFACE s_axilite port=numIters bundle=control"
            )

        self.code_gen_dict["$PRAGMAS$"].append("#pragma HLS INTERFACE ap_ctrl_none port=return")

    def get_number_output_values(self):
        return self.get_nodeattr("NumIters")

    def get_input_datatype(self, ind=0):
        # not supported
        raise Exception("get_input_datatype not implemented for TlastMarker")

    def get_output_datatype(self, ind=0):
        # not supported
        raise Exception("get_output_datatype not implemented for TlastMarker")

    def get_normal_input_shape(self, ind=0):
        # not supported
        raise Exception("get_normal_input_shape not implemented for TlastMarker")

    def get_normal_output_shape(self, ind=0):
        # not supported
        raise Exception("get_normal_input_shape not implemented for TlastMarker")

    def get_folded_input_shape(self, ind=0):
        stream_width = self.get_nodeattr("StreamWidth")
        elem_width = self.get_nodeattr("ElemWidth")
        n_packed_elems = stream_width // elem_width
        n_iters = self.get_nodeattr("NumIters")
        return (1, n_iters, n_packed_elems)

    def get_folded_output_shape(self, ind=0):
        return self.get_folded_input_shape()

    def get_instream_width(self, ind=0):
        stream_width = self.get_nodeattr("StreamWidth")
        return stream_width

    def get_outstream_width(self, ind=0):
        stream_width = self.get_nodeattr("StreamWidth")
        return stream_width

    def strm_decl(self):
        self.code_gen_dict["$STREAMDECLARATIONS$"] = []
        self.code_gen_dict["$STREAMDECLARATIONS$"].append('hls::stream<InDType> in0_V ("in0_V");')
        self.code_gen_dict["$STREAMDECLARATIONS$"].append(
            'hls::stream<OutDType> out0_V ("out0_V");'
        )

    def get_verilog_top_module_intf_names(self):
        intf_names = super().get_verilog_top_module_intf_names()
        stream_width = self.get_nodeattr("StreamWidth")
        intf_names["s_axis"] = [("in0_V", stream_width)]
        intf_names["m_axis"] = [("out0_V", stream_width)]
        if self.get_nodeattr("DynIters") == 1:
            intf_names["axilite"] = ["s_axi_control"]
        return intf_names



# Copyright (c) 2020, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.custom_op.registry import getCustomOp
from qonnx.transformation.base import Transformation
from qonnx.transformation.create_generic_partitions import PartitionFromLambda
from qonnx.util.basic import get_by_name

from finn.transformation.fpgadataflow.externalize_params import ExternalizeParams
from finn.util.basic import make_build_dir


class CreateDataflowPartition(Transformation):
    """Split a graph into two graphs; one which contains non-FINN-dataflow nodes
    and a StreamingDataflowPartition node, and another which only contains
    FINN dataflow nodes. The StreamingDataflowPartition has a model attribute
    that indicates the filename for the second graph that only contains
    dataflow nodes. No action is taken if there are no dataflow nodes."""

    def __init__(self, partition_model_dir=None):
        super().__init__()
        if partition_model_dir is None:
            self.partition_model_dir = make_build_dir("dataflow_partition_")
        else:
            self.partition_model_dir = partition_model_dir

    def apply(self, model):
        def filter_fc_extw(x):
            if x.op_type == "IODMA_hls":
                burst_mode = get_by_name(x.attribute, "burstMode")
                if burst_mode is not None:
                    burst_mode = burst_mode.s.decode("UTF-8")
                    return burst_mode == "wrap"

        extw_dma_nodes = list(filter(filter_fc_extw, model.graph.node))
        if len(extw_dma_nodes) > 0:
            model = model.transform(ExternalizeParams())

        def assign_partition_id(node):
            if node.op_type in ["GenericPartition", "StreamingDataflowPartition"]:
                return -1
            else:
                backend = get_by_name(node.attribute, "backend")
                if backend is not None and backend.s.decode("UTF-8") == "fpgadataflow":
                    assigned_partition = get_by_name(node.attribute, "partition_id")
                    if assigned_partition is not None:
                        return assigned_partition.i
                    else:
                        return 0
                else:
                    return -1

        # first, use the generic partitioning functionality to split up the graph
        parent_model = model.transform(
            PartitionFromLambda(
                partitioning=assign_partition_id, partition_dir=self.partition_model_dir
            )
        )
        # change node types to StreamingDataflowPartition
        p_nodes = parent_model.get_nodes_by_op_type("GenericPartition")
        for partition_ind, p_node in enumerate(p_nodes):
            # go into partition to extract some info
            p_node_inst = getCustomOp(p_node)
            node_model_filename = p_node_inst.get_nodeattr("model")
            p_model = ModelWrapper(node_model_filename)
            # check floorplan (SLR assignment per node)
            inst = getCustomOp(p_model.graph.node[0])
            slr = inst.get_nodeattr("slr")
            for node in p_model.graph.node:
                inst = getCustomOp(node)
                assert slr == inst.get_nodeattr(
                    "slr"
                ), """all nodes with same partition_id must have the same slr id"""
            # check that there is only one non-null mem_port per partition
            nmemports = 0
            mem_port = ""
            for node in p_model.graph.node:
                inst = getCustomOp(node)
                port = inst.get_nodeattr("mem_port")
                if port is not None and port != "":
                    nmemports += 1
                    mem_port = port
            assert nmemports <= 1, """Too many memory ports per partition"""
            # done, change node type and add info in parent graph
            p_node.op_type = "StreamingDataflowPartition"
            p_node.domain = "finn.custom_op.fpgadataflow"
            new_p_node_inst = getCustomOp(p_node)
            new_p_node_inst.set_nodeattr("partition_id", partition_ind)
            new_p_node_inst.set_nodeattr("slr", slr)
            new_p_node_inst.set_nodeattr("mem_port", mem_port)

        return (parent_model, False)



#############################################################################
# Copyright (C) 2025, Advanced Micro Devices, Inc.
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause
#
# @brief	SimEngine abstraction for running FINN task in simulated hardware.
# @author	Thomas B. Preußer <thomas.preusser@amd.com>
# @author	Yaman Umuroglu <yaman.umuroglu@amd.com>
#############################################################################

import xsi


class SimEngine:
    # ------------------------------------------------------------------------
    # Life Cycle
    def __init__(self, kernel, design, log=None, wdb=None):
        top = xsi.Design(xsi.Kernel(kernel), design, log, wdb)
        clk = top.getPort("ap_clk")
        clk2x = top.getPort("ap_clk2x")
        for p in top.ports():
            if p.isInput():
                p.clear().write_back()

        def cycle(updates):
            # Rising Edge
            clk.set(1).write_back()
            if clk2x is not None:
                clk2x.set(1).write_back()
            # Updates after Active Edge
            top.run(1)
            for port, update in updates.items():
                port.set_hexstr(update).write_back()

            # Edges inactive on interface & finish Cycle
            if clk2x is None:
                top.run(4999)
                clk.set(0).write_back()
                top.run(5000)
            else:
                top.run(2499)
                clk2x.set(0).write_back()
                top.run(2500)
                clk.set(0).write_back()
                clk2x.set(1).write_back()
                top.run(2500)
                clk2x.set(0).write_back()
                top.run(2500)

        self.top = top
        self.cycle = cycle
        self.ticks = 0
        self.tasks = []
        self.watchdogs = []

    # ------------------------------------------------------------------------
    # Utility
    def get_bus_port(self, bus, suffix):
        port = self.top.getPort(bus + "_" + suffix.lower())
        return port if port is not None else self.top.getPort(bus + "_" + suffix.upper())

    # ------------------------------------------------------------------------
    # Simulation Setup

    # Task Scheduling
    def enlist(self, task):
        self.tasks.append(task)

    # Watchdog Generation
    def create_watchdog(self, name, timeout):
        class Watchdog:
            def __init__(self, name, timeout):
                self.name = name
                self.ticks = 0
                self.timeout = timeout

            def __bool__(self):
                return self.ticks < self.timeout

            def __repr__(self):
                return self.name

            def __call__(self):
                self.ticks += 1

            def reset(self):
                self.ticks = 0

        ret = Watchdog(name, timeout)
        self.watchdogs.append(ret)
        return ret

    def remove_watchdog(self, watchdog):
        self.watchdogs.remove(watchdog)

    # ------------------------------------------------------------------------
    # Execution
    def run(self, cycles=None):
        "Run all tasks to completion or until a watchdog triggers."
        timeout = None if cycles is None else self.create_watchdog("Run Timeout", cycles)

        woken = []
        while len(self.tasks) > 0 and len(woken := [w for w in self.watchdogs if not w]) == 0:
            # Process Tasks and Collect Updates to Write Back
            tasks = []
            updates = {}

            # Execute Cycle
            self.ticks += 1
            strong = False
            for task in self.tasks:
                # Tasks read signals and derive updates to schedule for after the clock cycle
                ret = task(self)
                if ret is not None:
                    updates.update(ret)
                    tasks.append(task)
                    strong |= bool(task)
            self.cycle(updates)

            # Step Watchdogs
            for watchdog in self.watchdogs:
                watchdog()

            # Update to Unfinished Tasks
            self.tasks = tasks if strong else []

        # Return List of Woken Watchdogs
        if timeout is not None:
            self.remove_watchdog(timeout)
        return woken

    # ------------------------------------------------------------------------
    # Standard Tasks
    def do_reset(self):
        "Schedule a reset sequence."

        class Reset:
            def __init__(self, top):
                self.cnt = 0
                self.rst_n = top.getPort("ap_rst_n")

            def __call__(self, sim):
                cnt = self.cnt
                self.cnt = cnt + 1

                if cnt == 0:
                    return {self.rst_n: "0"}
                if cnt < 16:
                    return {}
                if cnt == 16:
                    return {self.rst_n: "1"}
                return None

        self.enlist(Reset(self.top))

    def stream_input(self, istream, values, throttle=(float("inf"), 0)):
        "Stream all values from the passed iterator into the specified stream."

        class InputStreamer:
            def __init__(self, top, istream, values, throttle):
                self.vld = top.get_bus_port(istream, "tvalid")
                self.rdy = top.get_bus_port(istream, "tready")
                self.dat = top.get_bus_port(istream, "tdata")
                self.values = values

                self.throttle = throttle
                self.await_tick = 0
                self.count_txns = throttle[0]

            def __call__(self, sim):
                vld = self.vld.as_bool()
                if vld and not self.rdy.read().as_bool():
                    return {}

                # Track Transaction Count
                if vld:
                    self.count_txns += 1

                # Proceed according to Throttling Rate
                if self.count_txns < self.throttle[0] or not sim.ticks < self.await_tick:
                    # Try Feed
                    val = next(self.values, None)
                    if val is None:
                        # Unset vld, then exit
                        return {self.vld: "0", self.dat: "0"} if vld else None

                    # Feed next Value
                    ret = {self.dat: val}
                    if not vld:
                        ret[self.vld] = "1"
                    if self.count_txns == self.throttle[0]:
                        self.count_txns = 0
                        self.await_tick = sim.ticks + self.throttle[1]
                    return ret

                # Stall Feed
                return {self.vld: "0", self.dat: "0"} if vld else {}

        self.enlist(InputStreamer(self, istream, values, throttle))

    def collect_output(self, ostream, size, watchdog=None):
        "Collect size outputs from the specified stream into the returned iterable buffer."

        class OutputCollector:
            def __init__(self, top, ostream, size, watchdog):
                self.size = size
                self.vld = top.get_bus_port(ostream, "tvalid")
                self.rdy = top.get_bus_port(ostream, "tready")
                self.dat = top.get_bus_port(ostream, "tdata")
                self.buf = []
                self.watchdog = watchdog

            def __iter__(self):
                return iter(self.buf)

            def __call__(self, sim):
                if self.rdy.as_bool():
                    if self.vld.read().as_bool():
                        # Have a n Output Transaction
                        if self.watchdog is not None:
                            self.watchdog.reset()
                        val = self.dat.read().as_hexstr()
                        self.buf.append(val)
                        if len(self.buf) == size:
                            return {self.rdy: "0"}
                    return {}

                if len(self.buf) < size:
                    return {self.rdy: "1"}
                return None

        ret = OutputCollector(self, ostream, size, watchdog)
        self.enlist(ret)
        return ret

    def trace_stream(self, stream):
        "Monitor an AXI-Stream and trace its transaction activity"

        class StreamTracer:
            def __init__(self, sim, stream):
                self.vld = sim.get_bus_port(stream, "tvalid")
                self.rdy = sim.get_bus_port(stream, "tready")
                self.trace = ""

            def __call__(self, sim):
                self.trace += (
                    "1" if self.vld.read().as_bool() and self.rdy.read().as_bool() else "0"
                )
                return {}

            def __bool__(self):
                return False

            def __str__(self):
                return self.trace

        ret = StreamTracer(self, stream)
        self.enlist(ret)
        return ret

    def write_axilite(self, m_axilite, writes):
        "Execute writes specified as a list of (addr, val)-tuples to AXI-lite interface"

        class AxiLiteWriter:
            INIT = 0
            FEED = 1
            COOL = 2

            def __init__(self, top, m_axilite, writes):
                self.awready = top.get_bus_port(m_axilite, "awready")
                self.awvalid = top.get_bus_port(m_axilite, "awvalid")
                self.awaddr = top.get_bus_port(m_axilite, "awaddr")
                self.wready = top.get_bus_port(m_axilite, "wready")
                self.wvalid = top.get_bus_port(m_axilite, "wvalid")
                self.wdata = top.get_bus_port(m_axilite, "wdata")
                wstrb = top.get_bus_port(m_axilite, "wstrb")
                wstrb.set_binstr("1" * wstrb.width()).write_back()
                self.bready = top.get_bus_port(m_axilite, "bready")
                self.bvalid = top.get_bus_port(m_axilite, "bvalid")
                self.bresp = top.get_bus_port(m_axilite, "bresp")
                self.writes = writes
                self.state = self.INIT
                self.pending = 0

            def __call__(self, sim):
                # Termination
                if self.state == self.COOL and not self.bready.as_bool():
                    return None

                ret = {}

                # Always Monitor Completions
                if self.state == self.INIT:
                    ret[self.bready] = "1"
                    self.state = self.FEED

                if self.bvalid.read().as_bool():
                    if self.pending < 1:
                        print("Received spurious completion on", self.bresp.name())
                    else:
                        self.pending -= 1
                        if self.pending == 0 and self.state == self.COOL:
                            ret[self.bready] = "0"

                    if self.bresp.read().as_unsigned() != 0:
                        print("Received error indication on", self.bresp.name())

                # Transaction Feed
                if self.state == self.FEED:
                    step = True

                    # Check for busy address feed
                    avld = self.awvalid.as_bool()
                    aclr = False
                    if avld:
                        if self.awready.read().as_bool():
                            aclr = True
                        else:
                            step = False

                    # Check for busy data feed
                    wvld = self.wvalid.as_bool()
                    wclr = False
                    if wvld:
                        if self.wready.read().as_bool():
                            wclr = True
                        else:
                            step = False

                    # Proceed with next Write
                    if step:
                        addr, val = next(self.writes, (None, None))
                        if addr is not None:
                            ret[self.awaddr] = f"{addr:x}"
                            ret[self.wdata] = val
                            if not avld:
                                ret[self.awvalid] = "1"
                            if not wvld:
                                ret[self.wvalid] = "1"
                            self.pending += 1
                            return ret
                        if not self.pending:
                            ret[self.bready] = "0"
                        self.state = self.COOL

                    # Deassert completed feed
                    if aclr:
                        ret[self.awvalid] = "0"
                    if wclr:
                        ret[self.wvalid] = "0"

                return ret

        self.enlist(AxiLiteWriter(self, m_axilite, writes))

    def read_axilite(self, m_axilite, reads):
        class AxiLiteReader:
            def __init__(self, top, m_axilite, reads):
                self.arready = top.get_bus_port(m_axilite, "arready")
                self.arvalid = top.get_bus_port(m_axilite, "arvalid")
                self.araddr = top.get_bus_port(m_axilite, "araddr")
                self.rready = top.get_bus_port(m_axilite, "rready")
                self.rvalid = top.get_bus_port(m_axilite, "rvalid")
                self.rdata = top.get_bus_port(m_axilite, "rdata")
                self.reads = reads
                self.pending = []
                self.draining = False
                self.replies = {}

            def __call__(self, sim):
                ret = {}

                # Address Stream Feed: assert self.draining when done
                if not self.draining:
                    if self.arready.read().as_bool() or not self.arvalid.as_bool():
                        addr = next(self.reads, None)
                        if addr is None:
                            ret[self.arvalid] = "0"
                            self.draining = True
                        else:
                            ret[self.arvalid] = "1"
                            ret[self.araddr] = f"{addr:x}"
                            self.pending.append(addr)

                # Reply Collection
                if not self.rready.as_bool():
                    # Termination
                    if self.draining:
                        return None
                    # Activation
                    ret[self.rready] = "1"
                elif self.rvalid.read().as_bool():
                    assert len(self.pending) > 0, "Spurious reply."
                    self.replies[self.pending.pop(0)] = self.rdata.read().as_hexstr()
                    if self.draining and len(self.pending) == 0:
                        ret[self.rready] = "0"

                return ret

            def __iter__(self):
                return iter(self.replies)

            def __getitem__(self, addr):
                return self.replies[addr]

        ret = AxiLiteReader(self, m_axilite, reads)
        self.enlist(ret)
        return ret
